import random
from collections import abc

from patent_client.util import Model

random.seed(1)


def compare_lists(list_1, list_2, key=""):
    for i, item in enumerate(list_1):
        if isinstance(item, list) and not isinstance(item, str):
            return compare_lists(item, list_2[i], key=f"{key}.{i}")
        elif isinstance(item, dict):
            return compare_dicts(item, list_2[i], key=f"{key}.{i}")
        else:
            assert item == list_2[i], f"At key {key}.{i}, {item} != {list_2[i]}"


def compare_dicts(dict_1, dict_2, key=""):
    dict_1_keys = set(dict_1.keys())
    dict_2_keys = set(dict_2.keys())
    in_1_not_2 = dict_1_keys - dict_2_keys
    in_2_not_1 = dict_2_keys - dict_1_keys
    if len(in_1_not_2) > 0:
        raise ValueError(f"At {key}, Keys {in_1_not_2} are in Dict 1, but not Dict 2")
    if len(in_2_not_1) > 0:
        raise ValueError(f"At {key}, Keys {in_2_not_1} are not in Dict 1, but are in Dict 2")

    for k, v in dict_1.items():
        if isinstance(v, list) and not isinstance(v, str):
            return compare_lists(v, dict_2[k], key=f"{key}.{k}")
        elif isinstance(v, dict):
            return compare_dicts(v, dict_2[k], key=f"{key}.{k}")
        else:
            assert v == dict_2[k], f"At key {key}.{k}, {v} != {dict_2[k]}"


def autogen_list_tests(name, lst):
    output = list()
    idx = random.randint(0, len(lst) - 1)
    subel_name = f"{name}[{idx}]"
    subel = lst[idx]
    output.append(f"assert len({name}) == {len(lst)}")
    if isinstance(subel, (abc.Mapping, Model)):
        output += autogen_dict_tests(subel_name, subel)
    elif isinstance(subel, abc.Sequence) and not isinstance(subel, str):
        output += autogen_list_tests(subel_name, subel)
    else:
        output.append(f"assert {subel_name} == {repr(subel)}")
    return output


def autogen_dict_tests(name, d):
    output = list()
    for k, v in d.items():
        subel_name = f"{name}.{k}"
        subel = v
        if isinstance(subel, (abc.Mapping, Model)):
            output += autogen_dict_tests(subel_name, subel)
        elif isinstance(subel, abc.Sequence) and not isinstance(subel, str):
            output += autogen_list_tests(subel_name, subel)
        else:
            output.append(f"assert {name}.{k} == {repr(v)}")
    return output


def autogen_tests(name, obj):
    if isinstance(obj, (abc.Mapping, Model)):
        output = autogen_dict_tests(name, obj)
    elif isinstance(obj, abc.Sequence) and not isinstance(obj, str):
        output = autogen_list_tests(name, obj)
    else:
        raise ValueError(f"Object is type {type(obj)}! Must be either a Model, abc.Mapping, or abc.Sequence")
    return "\n".join(output)



import pytest

from .api import PtabApi


class TestPtabApi:
    @pytest.mark.asyncio
    async def test_can_get_document(self):
        page = await PtabApi.get_documents(document_identifier="170680401")
        assert len(page.docs) == 1

    @pytest.mark.asyncio
    async def test_can_get_document_range(self):
        page = await PtabApi.get_documents(
            proceeding_number="IPR2022-00037", document_filing_date=("2023-01-01", "2023-09-15")
        )
        assert len(page.docs) == 25

    @pytest.mark.asyncio
    async def test_can_get_proceeding(self):
        page = await PtabApi.get_proceedings(proceeding_number="IPR2022-00037")
        assert len(page.docs) == 1



import datetime
from typing import List
from typing import Optional

from .model import PedsPage
from .schema import PedsPageSchema
from .session import session

type_map = {
    "string": str,
    "date": datetime.datetime,
    "text_general": str,
    "int": int,
    "text_ws": str,
}


class NotAvailableException(Exception):
    pass


class PatentExaminationDataSystemApi:
    base_url = "https://ped.uspto.gov/api"

    def __init__(self):
        query_id = None

    @classmethod
    async def is_online(cls) -> bool:
        response = await session.get("https://ped.uspto.gov/api/search-fields")
        if response.status_code == 200:
            return True
        elif "requested resource is not available" in response.text:
            raise NotAvailableException("Patent Examination Data is Offline - this is a USPTO problem")
        elif "attempt failed or the origin closed the connection" in response.text:
            raise NotAvailableException("The Patent Examination Data API is Broken! - this is a USPTO problem")
        else:
            raise NotAvailableException("There is a USPTO problem")

    @classmethod
    async def get_search_fields(cls) -> dict:
        if hasattr(cls, "search_fields"):
            return cls.search_fields
        response = await session.get("https://ped.uspto.gov/api/search-fields")
        response.raise_for_status()
        search_fields = response.json()
        cls.search_fields = {k: type_map[v] for k, v in search_fields.items()}
        return cls.search_fields

    async def create_query(
        self,
        query: str,
        query_fields: str = None,
        default_field: Optional[str] = "patentTitle",
        facet: str = False,
        return_fields: Optional[List[str]] = None,
        filter_query: Optional[List[str]] = None,
        minimum_match: str = "100%",
        sort: Optional[str] = "applId asc",
        start: int = 0,
        rows: Optional[int] = None,
    ) -> PedsPage:
        """

        Args:
            query (str): The query to be issued to PEDS
            query_fields (str): A list of fields to be queried
            default_field (str, optional): Default field to search. Defaults to None.
            facet (str, optional): Whether to enabled SOLR faceting. Defaults to "false".
            return_fields (str, optional): Specifies which fields should be returned. Defaults to "*".
            filter_query (str, optional)

        Returns:
            _type_: _description_
        """
        if query_fields is None:
            query_fields = list(f for f in (await self.get_search_fields()).keys() if f in query)
        if query_fields is None:
            query_fields = list((await self.get_search_fields()).keys())

        params = {
            "df": default_field,
            "qf": " ".join(query_fields),
            "fl": " ".join(return_fields) if return_fields else "*",
            "fq": filter_query if filter_query else list(),
            "searchText": query,
            "sort": sort,
            "facet": "true" if facet else "false",
            "mm": minimum_match,
            "start": str(start),
        }
        if rows:
            params["rows"] = rows

        url = "https://ped.uspto.gov/api/queries"
        response = await session.post(
            url,
            json=params,
            headers={"Accept": "application/json"},
        )
        response.raise_for_status()
        return PedsPageSchema().load(response.json())



from dataclasses import dataclass
from dataclasses import field

from patent_client.epo.ops.number_service.model import DocumentId
from patent_client.epo.ops.util import InpadocModel
from patent_client.util import Model
from yankee.data import ListCollection


@dataclass
class Citation(InpadocModel):
    cited_phase: str = None
    cited_by: str = None
    epodoc: DocumentId = None
    docdb: DocumentId = None
    original: DocumentId = None

    def __repr__(self):
        return f"Citation(doc_number={str(self.docdb)}, cited_by={self.cited_by}, cited_phase={self.cited_phase})"

    @property
    def docdb_number(self):
        return str(self.docdb)


@dataclass
class Title(Model):
    lang: str
    text: str

    def __repr__(self):
        return f"Title(lang={self.lang}, text={self.text})"


def limit_text(string, limit=30):
    if len(string) < limit:
        return string
    else:
        return f"{string[:limit]}..."


@dataclass
class InpadocBiblio(InpadocModel):
    __manager__ = "patent_client.epo.ops.published.manager.BiblioManager"
    country: str = None
    doc_number: str = None
    kind: str = None
    family_id: str = None
    publication_number: str = None
    application_number: str = None
    publication_reference_docdb: DocumentId = None
    publication_reference_epodoc: DocumentId = None
    publication_reference_original: DocumentId = None
    application_reference_docdb: DocumentId = None
    application_reference_epodoc: DocumentId = None
    application_reference_original: DocumentId = None
    intl_class: "ListCollection[str]" = field(default_factory=list)
    cpc_class: "ListCollection[str]" = field(default_factory=list)
    us_class: "ListCollection[str]" = field(default_factory=list)
    priority_claims: "ListCollection[str]" = field(default_factory=list)
    title: str = None
    titles: "ListCollection[Title]" = field(default_factory=list)
    abstract: str = None
    citations: "ListCollection[Citation]" = field(default_factory=list)
    applicants_epodoc: "ListCollection[str]" = field(default_factory=list)
    applicants_original: "ListCollection[str]" = field(default_factory=list)
    inventors_epodoc: "ListCollection[str]" = field(default_factory=list)
    inventors_original: "ListCollection[str]" = field(default_factory=list)
    # TODO: NPL citations

    def __repr__(self):
        return f"InpadocBiblio(publication_number={self.publication_number}, title={limit_text(self.title, 30)})"

    @property
    def docdb_number(self):
        return str(self.publication_reference_docdb)


@dataclass
class BiblioResult(Model):
    documents: list = None



import textwrap
from dataclasses import dataclass
from dataclasses import field
from typing import *

from ..base.model import Model


@dataclass
class Claim(Model):
    __exclude__ = ["depends_on_claim", "dependent_claims"]

    number: int
    # text: str
    limitations: "List[str]"
    depends_on: "List[int]" = field(default_factory=list)
    dependent_claims: "List[int]" = field(default_factory=list)

    def __repr__(self):
        return f"Claim(number={self.number}, depends_on={self.depends_on}, text={textwrap.shorten(self.text, width=40, placeholder='...')})"

    @property
    def text(self):
        return f"{self.number}. " + "\n".join(self.limitations)

    @property
    def independent(self):
        return not bool(self.depends_on)

    @property
    def dependent(self):
        return bool(self.depends_on)



import datetime
from dataclasses import dataclass
from dataclasses import field

from patent_client.epo.ops.number_service.model import DocumentId
from patent_client.util import Model
from yankee.data import ListCollection


@dataclass
class MetaData(Model):
    status_of_data: str = None
    docdb_publication_number: str = None
    subscriber_exchange_date: "datetime.date" = None
    epo_created_date: "datetime.date" = None
    docdb_integer: int = None


@dataclass
class LegalEvent(Model):
    """
    Field descriptions are here:
    https://documents.epo.org/projects/babylon/eponot.nsf/0/EF223017D933B30AC1257B50005A042E/$File/14.11_User_Documentation_3.1_en.pdf

    """

    filing_or_publication: str = None
    document_number: str = None
    ip_type: str = None
    metadata: MetaData = None
    country_code: str = None
    text_record: str = None

    event_date: "datetime.date" = None
    event_code: str = None
    event_country: str = None
    event_description: str = None
    regional_event_code: str = None

    corresponding_patent: str = None
    corresponding_patent_publication_date: "datetime.date" = None

    designated_states: str = None
    extension_name: str = None
    new_owner_name: str = None
    free_text: str = None
    spc_number: str = None
    spc_filing_date: str = None
    expiry: str = None
    publication_language: str = None
    inventor_name: str = None
    ipc_class: str = None
    representative_name: str = None
    payment_date: str = None
    opponent_name: str = None
    year_of_fee_payment: str = None
    name_of_requester: str = None
    date_extension_granted: str = None
    extension_states: str = None
    effective_date: "datetime.date" = None
    date_of_withdrawal: str = None

    def __repr__(self):
        return f"Event(document_number={self.document_number}, event_description={self.event_description}, event_date={self.event_date})"


@dataclass
class Legal(Model):
    __manager__ = "patent_client.epo.ops.legal.manager.LegalManager"
    publication_reference: "DocumentId" = None
    events: list = field(default_factory=ListCollection)



SEARCH_FIELDS = {
    "title": "title",
    "abstract": "abstract",
    "title_and_abstract": "titleandabstract",
    "inventor": "inventor",
    "applicant": "applicant",
    "inventor_or_applicant": "inventorandapplicant",
    "publication": "publicationnumber",
    "epodoc_publication": "spn",
    "application": "applicationnumber",
    "epodoc_application": "sap",
    "priority": "prioritynumber",
    "epodoc_priority": "spr",
    "number": "num",  # Pub, App, or Priority Number
    "publication_date": "publicationdate",  # yyyy, yyyyMM, yyyyMMdd, yyyy-MM, yyyy-MM-dd
    "citation": "citation",
    "cited_in_examination": "ex",
    "cited_in_opposition": "op",
    "cited_by_applicant": "rf",
    "other_citation": "oc",
    "family": "famn",
    "cpc_class": "cpc",
    "ipc_class": "ipc",
    "ipc_core_invention_class": "ci",
    "ipc_core_additional_class": "cn",
    "ipc_advanced_class": "ai",
    "ipc_advanced_additional_class": "an",
    "ipc_core_class": "c",
    "classification": "cl",  # IPC or CPC Class
    "full_text": "txt",  # title, abstract, inventor and applicant
}


def generate_query(**kwargs):
    query = list()
    for keyword, values in kwargs.items():
        if isinstance(values, list):
            for value in values:
                if keyword:
                    query.append(f'{SEARCH_FIELDS[keyword]}="{value}"')
        else:
            if keyword:
                query.append(f'{SEARCH_FIELDS[keyword]}="{values}"')
    return " AND ".join(query)



import asyncio


def run_sync(coroutine):
    return asyncio.run(coroutine)


def run_async_iterator(async_iterator):
    while True:
        try:
            yield run_sync(async_iterator.__anext__())
        except StopAsyncIteration:
            break


def run_sync_decorator(func):
    def wrapper(*args, **kwargs):
        return run_sync(func(*args, **kwargs))

    return wrapper


def run_sync_iterator_decorator(func):
    def wrapper(*args, **kwargs):
        return run_async_iterator(func(*args, **kwargs))

    return wrapper


class SyncProxy:
    def __init__(self, async_obj):
        self.async_obj = async_obj

    def __getattr__(self, name):
        return run_sync_decorator(getattr(self.async_obj, name))

    def __iter__(self):
        return run_sync_iterator_decorator(self.async_obj.__aiter__())



from dataclasses import dataclass

import pytest

from .model import Model


def test_model():
    class DummyManager:
        pass

    @dataclass
    class Example(Model):
        __manager__ = DummyManager
        a: str = None
        b: int = None

    ex = Example(a="1", b=2)
    assert len(ex.fields()) == 2
    assert ex.a == "1"
    assert ex.b == 2
    d = ex.to_dict(item_class=dict)
    assert type(d) == dict
    with pytest.raises(AttributeError):
        assert d.a == "1"
    assert d == {"a": "1", "b": 2}


def test_exclude_fields():
    @dataclass
    class Example(Model):
        __exclude_fields__ = [
            "b",
        ]
        a: str = None
        b: int = None

    ex = Example(a="1", b=2)
    row = ex.to_dict()
    assert "b" not in row



from patent_client.util.base.manager import Manager

from .api import GlobalDossierAsyncApi
from .query import QueryBuilder
from .schema import DocumentListSchema
from .schema import GlobalDossierSchema

query_builder = QueryBuilder()

global_dossier_async_api = GlobalDossierAsyncApi()


class GlobalDossierBaseManager(Manager):
    def filter(self, *args, **kwargs):
        raise NotImplementedError("GlobalDossier can only retrieve using the GET interface")

    def order_by(self, *args, **kwargs):
        raise NotImplementedError("GlobalDossier can only retrieve using the GET interface")

    def limit(self, *args, **kwargs):
        raise NotImplementedError("GlobalDossier can only retrieve using the GET interface")

    def offset(self, *args, **kwargs):
        raise NotImplementedError("GlobalDossier can only retrieve using the GET interface")


class GlobalDossierManager(GlobalDossierBaseManager):
    __schema__ = GlobalDossierSchema

    async def aget(self, *args, **kwargs):
        data = await global_dossier_async_api.get_file(**query_builder.build_query(*args, **kwargs))
        return self.__schema__.load(data)


class GlobalDossierApplicationManager(GlobalDossierBaseManager):
    __schema__ = GlobalDossierSchema

    async def aget(self, *args, **kwargs):
        query = query_builder.build_query(*args, **kwargs)
        data = await global_dossier_async_api.get_file(**query)
        gd_file = self.__schema__.load(data)
        if query["type_code"] == "application":
            return next(a for a in gd_file.applications if a.app_num in query["doc_number"])
        elif query["type_code"] == "publication":
            return next(a for a in gd_file.applications if any(p.pub_num in query["doc_number"] for p in a.pub_list))


class GlobalDossierDocument(GlobalDossierBaseManager):
    __schema__ = DocumentListSchema

    async def aget(self, country, doc_number, kind_code):
        data = await global_dossier_async_api.get_doc_list(country, doc_number, kind_code)
        return self.__schema__.load(data)



from dataclasses import dataclass
from dataclasses import field
from pathlib import Path
from typing import List

from patent_client.epo.ops.util import InpadocModel
from patent_client.util import Model
from PyPDF2 import PdfReader
from PyPDF2 import PdfWriter


@dataclass
class Section(Model):
    name: str = None
    start_page: int = None


@dataclass
class ImageDocument(Model):
    num_pages: int = None
    description: str = None
    link: str = None
    formats: List[str] = field(default_factory=list)
    sections: List[Section] = field(default_factory=list)
    doc_number: str = None

    def download(self, path="."):
        from ..api import PublishedImagesApi

        out_file = Path(path) / f"{self.doc_number}.pdf"
        writer = PdfWriter()
        for i in range(1, self.num_pages + 1):
            page_data = PublishedImagesApi.get_page_image_from_link(self.link, page_number=i)
            page = PdfReader(page_data).pages[0]
            if page["/Rotate"] == 90:
                page.rotate_clockwise(-90)
            writer.add_page(page)

        for section in self.sections:
            writer.add_outline_item(section.name.capitalize(), section.start_page)

        with out_file.open("wb") as f:
            writer.write(f)

    def download_image(self, path=".", image_format="tif", page_number=1):
        from ..api import PublishedImagesApi
        
        out_file = Path(path) / f"{self.doc_number}.{image_format}"
        
        image = PublishedImagesApi.get_page_image_from_link(self.link, page_number=page_number, image_format=image_format) 
        
        with out_file.open("wb") as f:
            f.write(image.read())


@dataclass
class Images(InpadocModel):
    __manager__ = "patent_client.epo.ops.published.manager.ImageManager"
    publication_number: str = None
    full_document: ImageDocument = None
    drawing: ImageDocument = None
    first_page: ImageDocument = None

    @property
    def docdb_number(self):
        return str(self.publication_reference)



from patent_client.parser import parse


class TestPatentNumberParser:
    def test_can_handle_integer_values(self):
        pat = parse(6013599)
        assert pat.number == "6013599"
        assert pat.country == "US"
        assert pat.kind_code == "B2"
        assert pat.display() == "US 6,013,599 B2"
        assert str(pat) == "US6013599B2"

        app = parse(20150012345)
        assert app.number == "20150012345"
        assert app.country == "US"
        assert app.kind_code == "A1"
        assert app.display() == "US 2015/0012345 A1"
        assert str(app) == "US20150012345A1"

    def test_can_handle_string_values_without_country_codes(self):
        pat = parse("6013599")
        assert pat.number == "6013599"
        assert pat.country == "US"
        assert pat.kind_code == "B2"
        assert pat.display() == "US 6,013,599 B2"
        assert str(pat) == "US6013599B2"

        app = parse("20150012345")
        assert app.number == "20150012345"
        assert app.country == "US"
        assert app.kind_code == "A1"
        assert app.display() == "US 2015/0012345 A1"
        assert str(app) == "US20150012345A1"

    def test_can_handle_us_without_kind_codes(self):
        pat = parse("US6013599")
        assert pat.number == "6013599"
        assert pat.country == "US"
        assert pat.kind_code == "B2"
        assert pat.display() == "US 6,013,599 B2"
        assert str(pat) == "US6013599B2"

        app = parse("US20150012345")
        assert app.number == "20150012345"
        assert app.country == "US"
        assert app.kind_code == "A1"
        assert app.display() == "US 2015/0012345 A1"
        assert str(app) == "US20150012345A1"

    def test_can_handle_us_with_kind_codes(self):
        pat = parse("US6013599B2")
        assert pat.number == "6013599"
        assert pat.country == "US"
        assert pat.kind_code == "B2"
        assert pat.display() == "US 6,013,599 B2"
        assert str(pat) == "US6013599B2"

        app = parse("US20150012345A1")
        assert app.number == "20150012345"
        assert app.country == "US"
        assert app.kind_code == "A1"
        assert app.display() == "US 2015/0012345 A1"
        assert str(app) == "US20150012345A1"

    def test_can_handle_us_in_display_form(self):
        pat = parse("US 6,013,599 B2")
        assert pat.number == "6013599"
        assert pat.country == "US"
        assert pat.kind_code == "B2"
        assert pat.display() == "US 6,013,599 B2"
        assert str(pat) == "US6013599B2"

        app = parse("US 2015/0012345 A1")
        assert app.number == "20150012345"
        assert app.country == "US"
        assert app.kind_code == "A1"
        assert app.display() == "US 2015/0012345 A1"
        assert str(app) == "US20150012345A1"

    def test_can_handle_us_design_patents(self):
        app = parse("USD645062")
        assert app.number == "D645062"
        assert app.country == "US"
        assert app.display() == "US D645062"
        assert str(app) == "USD645062"

    def test_can_handle_us_application_number(self):
        app = parse("14123456")
        assert app.number == "14123456"
        assert app.country == "US"
        assert app.kind_code == ""
        assert app.display() == "US 14/123,456"
        assert str(app) == "US14123456"

        app = parse("US14123456")
        assert app.number == "14123456"
        assert app.country == "US"
        assert app.kind_code == ""
        assert app.display() == "US 14/123,456"
        assert str(app) == "US14123456"

        app = parse("14/123,456")
        assert app.number == "14123456"
        assert app.country == "US"
        assert app.kind_code == ""
        assert app.display() == "US 14/123,456"
        assert str(app) == "US14123456"

        app = parse("14/123456")
        assert app.number == "14123456"
        assert app.country == "US"
        assert app.kind_code == ""
        assert app.display() == "US 14/123,456"
        assert str(app) == "US14123456"

        app = parse("US 14/123,456")
        assert app.number == "14123456"
        assert app.country == "US"
        assert app.kind_code == ""
        assert app.display() == "US 14/123,456"
        assert str(app) == "US14123456"

        app = parse(14123456)
        assert app.number == "14123456"
        assert app.country == "US"
        assert app.kind_code == ""
        assert app.display() == "US 14/123,456"
        assert str(app) == "US14123456"

    def test_can_handle_us_application_number_with_leading_zeroes(self):
        app = parse("09054439")
        assert app.number == "09054439"
        assert app.country == "US"
        assert app.kind_code == ""
        assert app.display() == "US 09/054,439"
        assert str(app) == "US09054439"

    def test_can_handle_reexam_cases(self):
        app = parse(95000486)
        assert app.number == "95000486"
        assert app.country == "US"
        assert app.kind_code == ""
        assert app.display() == "US 95/000,486"
        assert str(app) == "US95000486"

    def test_can_handle_reissue_applications(self):
        app = parse("RE43633")
        assert app.country == "US"
        assert app.number == "RE43633"
        assert app.kind_code == "E1"
        assert app.display() == "US RE43633 E1"
        assert str(app) == "USRE43633E1"

        app = parse("USRE43633")
        assert app.country == "US"
        assert app.number == "RE43633"
        assert app.kind_code == "E1"
        assert app.display() == "US RE43633 E1"
        assert str(app) == "USRE43633E1"

    def test_can_handle_pct_applications(self):
        app = parse("PCT/US17/36577")
        assert app.type == "international application"
        assert app.country == "US"
        assert app.number == "036577"
        assert app.year == "2017"
        assert app.kind_code == ""
        assert app.display() == "PCT/US17/36577"
        assert str(app) == "PCTUS2017036577"

        app = parse("PCT/US2017/036577")
        assert app.type == "international application"
        assert app.country == "US"
        assert app.number == "036577"
        assert app.year == "2017"
        assert app.kind_code == ""
        assert app.display(style="new") == "PCT/US2017/036577"
        assert app.display() == "PCT/US17/36577"
        assert str(app) == "PCTUS2017036577"

    def test_can_handle_wipo_publications(self):
        app = parse("WO2009029879")
        # assert app.type == "international publication"
        assert app.country == "WO"
        assert app.number == "2009029879"
        assert app.kind_code == ""
        assert app.display() == "WO 2009029879"
        assert str(app) == "WO2009029879"

    def test_can_handle_canadian_patents(self):
        app = parse("CA2967774A")
        assert app.type == "application"
        assert app.country == "CA"
        assert app.number == "2967774"
        assert app.kind_code == "A"
        assert app.display() == "CA 2967774 A"
        assert str(app) == "CA2967774A"

        app = parse("2967774", country="CA")
        assert app.type == "application"
        assert app.country == "CA"
        assert app.number == "2967774"
        assert app.kind_code == "A"
        assert app.display() == "CA 2967774 A"
        assert str(app) == "CA2967774A"

        app = parse("CA2967774A1")
        assert app.country == "CA"
        assert app.number == "2967774"
        assert app.kind_code == "A1"
        assert app.type == "publication"
        assert app.display() == "CA 2967774 A1"
        assert str(app) == "CA2967774A1"

        app = parse("CA2967774E")
        assert app.country == "CA"
        assert app.number == "2967774"
        assert app.kind_code == "E"
        assert app.type == "reissue patent"
        assert app.display() == "CA 2967774 E"
        assert str(app) == "CA2967774E"

    def test_can_handle_8_digit_patent_numbers(self):
        pat = parse("11123123")
        assert pat.country == "US"
        assert pat.number == "11123123"
        assert pat.kind_code == "B2"
        assert pat.type == "patent"
        assert pat.display() == "US 11,123,123 B2"
        assert str(pat) == "US11123123B2"



import hishel
import httpx

session = hishel.AsyncCacheClient(
    headers={
        "X-Requested-With": "XMLHttpRequest",
        "User-Agent": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/114.0.0.0 Safari/537.36",
    },
    transport=httpx.AsyncHTTPTransport(
        retries=3,
        verify=False,
        http2=True,
    ),
    timeout=5 * 60,
)



import pytest

from .query import QueryBuilder
from .query import QueryException

builder = QueryBuilder()


def test_garbage():
    with pytest.raises(QueryException):
        builder.build_query("garbage")


def test_default():
    assert builder.build_query("12123456") == {
        "doc_number": "12123456",
        "type_code": "application",
        "office_code": "US",
    }


def test_us_string_queries():
    assert builder.build_query("RE12345") == {
        "doc_number": "RE12345",
        "type_code": "patent",
        "office_code": "US",
    }
    assert builder.build_query("1234567") == {
        "doc_number": "1234567",
        "type_code": "patent",
        "office_code": "US",
    }
    assert builder.build_query("1234567123456") == {
        "doc_number": "1234567123456",
        "type_code": "patent",
        "office_code": "US",
    }


def test_us_string_queries_with_country():
    assert builder.build_query("USRE12345") == {
        "doc_number": "RE12345",
        "type_code": "patent",
        "office_code": "US",
    }
    assert builder.build_query("US1234567") == {
        "doc_number": "1234567",
        "type_code": "patent",
        "office_code": "US",
    }
    assert builder.build_query("US1234567123456") == {
        "doc_number": "1234567123456",
        "type_code": "patent",
        "office_code": "US",
    }


def test_us_string_queries_with_defined_type():
    assert builder.build_query("US16123456", type="application") == {
        "doc_number": "16123456",
        "type_code": "application",
        "office_code": "US",
    }
    assert builder.build_query("US16123456", type="patent") == {
        "doc_number": "16123456",
        "type_code": "patent",
        "office_code": "US",
    }


def test_indefinite_us_string_queries():
    with pytest.raises(QueryException):
        builder.build_query("US16123456")
    with pytest.raises(QueryException):
        builder.build_query("12345678910")


def test_wipo_queries():
    assert builder.build_query("PCT/US15/12345") == {
        "doc_number": "PCT/US15/12345",
        "type_code": "application",
        "office_code": "WIPO",
    }
    assert builder.build_query("WO1234567") == {
        "doc_number": "WO1234567",
        "type_code": "publication",
        "office_code": "WIPO",
    }
    assert builder.build_query(application="PCT/US15/12345") == {
        "doc_number": "PCT/US15/12345",
        "type_code": "application",
        "office_code": "WIPO",
    }
    assert builder.build_query(publication="WO1234567") == {
        "doc_number": "WO1234567",
        "type_code": "publication",
        "office_code": "WIPO",
    }
    with pytest.raises(QueryException):
        assert builder.build_query(publication="PCT/US15/12345") == {
            "doc_number": "PCT/US15/12345",
            "type_code": "application",
            "office_code": "WIPO",
        }
    with pytest.raises(QueryException):
        assert builder.build_query(application="WO1234567") == {
            "doc_number": "WO1234567",
            "type_code": "publication",
            "office_code": "WIPO",
        }


def test_au_queries():
    with pytest.raises(QueryException):
        builder.build_query("AU123456")
    assert builder.build_query("AU123456", type="publication") == {
        "doc_number": "AU123456",
        "type_code": "publication",
        "office_code": "CASE",
    }
    assert builder.build_query("AU123456", type="application") == {
        "doc_number": "AU123456",
        "type_code": "application",
        "office_code": "CASE",
    }
    assert builder.build_query(application="AU123456") == {
        "doc_number": "AU123456",
        "type_code": "application",
        "office_code": "CASE",
    }
    assert builder.build_query(publication="AU123456") == {
        "doc_number": "AU123456",
        "type_code": "publication",
        "office_code": "CASE",
    }


def test_ep_queries():
    assert builder.build_query("EP1234567") == {
        "office_code": "EP",
        "doc_number": "1234567",
        "type_code": "publication",
    }
    assert builder.build_query("1234567", office="EP") == {
        "office_code": "EP",
        "doc_number": "1234567",
        "type_code": "publication",
    }
    assert builder.build_query(publication="EP1234567") == {
        "office_code": "EP",
        "doc_number": "1234567",
        "type_code": "publication",
    }
    assert builder.build_query(publication="1234567", office="EP") == {
        "office_code": "EP",
        "doc_number": "1234567",
        "type_code": "publication",
    }

    assert builder.build_query("EP12345678") == {
        "office_code": "EP",
        "doc_number": "12345678",
        "type_code": "application",
    }
    assert builder.build_query("12345678", office="EP") == {
        "office_code": "EP",
        "doc_number": "12345678",
        "type_code": "application",
    }
    assert builder.build_query(application="EP12345678") == {
        "office_code": "EP",
        "doc_number": "12345678",
        "type_code": "application",
    }
    assert builder.build_query(application="12345678", office="EP") == {
        "office_code": "EP",
        "doc_number": "12345678",
        "type_code": "application",
    }



import json
from pathlib import Path

from patent_client.util.test import compare_dicts

from .schema import PublicSearchDocumentSchema
from .schema import PublicSearchSchema

test_dir = Path(__file__).parent / "test"


class TestSchema:
    def test_biblio(self):
        data = (test_dir / "biblio.json").read_text()
        parser = PublicSearchSchema()
        result = parser.load_batch(data)
        expected_file = test_dir / "biblio_expected.json"
        expected_file.write_text(result.to_json(indent=2))
        expected = json.loads(expected_file.read_text())
        result_json = json.loads(result.to_json())
        for i in range(len(expected)):
            compare_dicts(expected[i], result_json[i])

    def test_docs(self):
        data = (test_dir / "docs.json").read_text()
        parser = PublicSearchDocumentSchema()
        result = parser.load_batch(data)
        expected_file = test_dir / "docs_expected.json"
        expected_file.write_text(result.to_json(indent=2))
        expected = json.loads(expected_file.read_text())
        result_json = json.loads(result.to_json())
        for i in range(len(expected)):
            compare_dicts(expected[i], result_json[i])



import hishel
import httpx

session = hishel.AsyncCacheClient(
    transport=httpx.AsyncHTTPTransport(
        retries=3,
        verify=False,
    ),
    timeout=60 * 5,
    controller=hishel.Controller(
        allow_heuristics=True,
    ),
)



from dataclasses import dataclass
from dataclasses import field
from typing import List

from patent_client.epo.ops.number_service.model import DocumentId
from patent_client.util import Model


@dataclass
class PriorityClaim(Model):
    application_number: str = None
    application_reference: DocumentId = None
    sequence: int = None
    kind: str = None
    active: bool = None


@dataclass
class FamilyMember(Model):
    publication_number: str = None
    application_number: str = None
    family_id: str = None
    publication_reference: list = field(default_factory=list)
    application_reference: list = field(default_factory=list)
    priority_claims: List[PriorityClaim] = field(default_factory=list)

    @property
    def docdb_number(self):
        return self.publication_number

    def __repr__(self):
        return f"FamilyMember(publication_number={self.publication_number})"


@dataclass
class Family(Model):
    __manager__ = "patent_client.epo.ops.family.manager.FamilyManager"
    publication_reference: DocumentId = None
    num_records: int = None
    publication_number: str = None
    family_members: List[FamilyMember] = field(default_factory=list)



import asyncio
from pathlib import Path

import httpx
import tenacity

from .session import session


class UsptoException(Exception):
    pass


def force_list(obj):
    if not isinstance(obj, list):
        return [
            obj,
        ]
    return obj


class PublicSearchAsyncApi:
    def __init__(self):
        self.session = dict()
        self.case_id = None

    @tenacity.retry(stop=tenacity.stop_after_attempt(3), wait=tenacity.wait_random(min=1, max=5))
    async def run_query(
        self,
        query,
        start=0,
        limit=500,
        sort="date_publ desc",
        default_operator="OR",
        sources=["US-PGPUB", "USPAT", "USOCR"],
        expand_plurals=True,
        british_equivalents=True,
    ):
        if self.case_id is None:
            await self.get_session()
        url = "https://ppubs.uspto.gov/dirsearch-public/searches/searchWithBeFamily"
        data = {
            "start": start,
            "pageCount": limit,
            "sort": sort,
            "docFamilyFiltering": "familyIdFiltering",
            "searchType": 1,
            "familyIdEnglishOnly": True,
            "familyIdFirstPreferred": "US-PGPUB",
            "familyIdSecondPreferred": "USPAT",
            "familyIdThirdPreferred": "FPRS",
            "showDocPerFamilyPref": "showEnglish",
            "queryId": 0,
            "tagDocSearch": False,
            "query": {
                "caseId": self.case_id,
                "hl_snippets": "2",
                "op": default_operator,
                "q": query,
                "queryName": query,
                "highlights": "1",
                "qt": "brs",
                "spellCheck": False,
                "viewName": "tile",
                "plurals": expand_plurals,
                "britishEquivalents": british_equivalents,
                "databaseFilters": [],
                "searchType": 1,
                "ignorePersist": True,
                "userEnteredQuery": query,
            },
        }
        for s in force_list(sources):
            data["query"]["databaseFilters"].append({"databaseName": s, "countryCodes": []})
        query_response = await session.post(url, json=data)
        if query_response.status_code in (500, 415):
            await asyncio.sleep(5)
            query_response = await session.post(url, json=data)
        query_response.raise_for_status()
        result = query_response.json()
        if result.get("error", None) is not None:
            raise UsptoException(f"Error #{result['error']['errorCode']}\n{result['error']['errorMessage']}")
        return result

    @tenacity.retry(stop=tenacity.stop_after_attempt(3), wait=tenacity.wait_random(min=1, max=5))
    async def get_document(self, bib):
        url = f"https://ppubs.uspto.gov/dirsearch-public/patents/{bib.guid}/highlight"
        params = {
            "queryId": 1,
            "source": bib.type,
            "includeSections": True,
            "uniqueId": None,
        }
        response = await session.get(url, params=params)
        response.raise_for_status()
        return response.json()

    @tenacity.retry(stop=tenacity.stop_after_attempt(3), wait=tenacity.wait_random(min=1, max=5))
    async def get_session(self):
        url = "https://ppubs.uspto.gov/dirsearch-public/users/me/session"
        response = await session.post(url, json=-1)  # json=str(random.randint(10000, 99999)))
        self.session = response.json()
        self.case_id = self.session["userCase"]["caseId"]
        return self.session

    @tenacity.retry(stop=tenacity.stop_after_attempt(3), wait=tenacity.wait_random(min=1, max=5))
    async def _request_save(self, obj):
        page_keys = [f"{obj.image_location}/{i:0>8}.tif" for i in range(1, obj.document_structure.page_count + 1)]
        response = await session.post(
            "https://ppubs.uspto.gov/dirsearch-public/print/imageviewer",
            json={
                "caseId": self.case_id,
                "pageKeys": page_keys,
                "patentGuid": obj.guid,
                "saveOrPrint": "save",
                "source": obj.type,
            },
        )
        response.raise_for_status()
        return response.text

    @tenacity.retry(stop=tenacity.stop_after_attempt(3), wait=tenacity.wait_random(min=1, max=5))
    async def download_image(self, obj, path="."):
        print("Trying!")
        out_path = Path(path).expanduser() / f"{obj.guid}.pdf"
        if out_path.exists():
            return out_path
        if self.case_id is None:
            await self.get_session()
        try:
            print_job_id = await self._request_save(obj)
        except httpx.HTTPStatusError:
            await self.get_session()
            print_job_id = await self._request_save(obj)
        while True:
            response = await session.post(
                "https://ppubs.uspto.gov/dirsearch-public/print/print-process",
                json=[
                    print_job_id,
                ],
            )
            response.raise_for_status()
            print_data = response.json()
            if print_data[0]["printStatus"] == "COMPLETED":
                break
            await asyncio.sleep(1)
        pdf_name = print_data[0]["pdfName"]
        with out_path.open("wb") as f:
            try:
                request = session.build_request(
                    "GET", f"https://ppubs.uspto.gov/dirsearch-public/print/save/{pdf_name}"
                )
                response = await session.send(request, stream=True)
                response.raise_for_status()
                async for chunk in response.aiter_bytes():
                    if chunk:
                        f.write(chunk)
            except httpx.HTTPStatusError as e:
                await response.aclose()
                raise e
        return out_path



from .assignment import Assignment
from .peds import USApplication
from .ptab import PtabDecision
from .ptab import PtabDocument
from .ptab import PtabProceeding

__all__ = [
    "Assignment",
    "USApplication",
    "PtabDecision",
    "PtabDocument",
    "PtabProceeding",
]



from typing import AsyncIterator
from typing import Iterator

from patent_client.util import Manager

from .api import PublishedAsyncApi
from .cql import generate_query
from .model.biblio import BiblioResult
from .model.fulltext import Claims
from .model.fulltext import Description
from .model.images import ImageDocument
from .schema import BiblioResultSchema
from .schema import ClaimsSchema
from .schema import DescriptionSchema
from .schema import ImagesSchema
from .schema import SearchSchema


class SearchManager(Manager):
    result_size = 100
    primary_key = "publication"
    __schema__ = SearchSchema
    __item_schema__ = BiblioResultSchema

    def __init__(self, config=None):
        super().__init__(config=None)
        if callable(self.__item_schema__):
            self.__item_schema__ = self.__item_schema__()

    def __iter__(self) -> Iterator[BiblioResult]:
        return super(SearchManager, self).__iter__()

    def __aiter__(self) -> AsyncIterator[BiblioResult]:
        return super(SearchManager, self).__aiter__()

    async def _aget_search_results_range(self, start=1, end=100):
        if "cql_query" in self.config.filter:
            query = self.config.filter["cql_query"]
        else:
            query = generate_query(**self.config.filter)
        return self.__schema__.load(await PublishedAsyncApi.search.search(query, start, end))

    async def alen(self) -> int:
        page = await self._aget_search_results_range(1, 100)
        offset = self.config.offset or 0
        limit = self.config.limit or page.num_results - offset
        num_results = page.num_results
        num_results -= offset
        num_results = min(limit, num_results)
        return num_results

    async def _aget_results(self):
        if len(self) == 0:
            return
        num_pages = round(len(self) / self.result_size)
        limit = self.config.limit or len(self)
        offset = self.config.offset or 0
        max_position = offset + limit
        range = (offset + 1, min(offset + self.result_size, max_position))
        while True:
            page = await self._aget_search_results_range(*range)
            for result in page.results:
                yield result
            if range[1] == max_position:
                break
            range = (
                range[0] + self.result_size,
                min(range[1] + self.result_size, max_position),
            )

    async def aget(self, number, doc_type="publication", format="docdb") -> BiblioResult:
        result = self.__item_schema__.load(await PublishedAsyncApi.biblio.get_biblio(number, doc_type, format))
        if len(result.documents) > 1:
            raise Exception("More than one result found! Try another query")
        return result.documents[0]


class BiblioManager(Manager):
    __schema__ = BiblioResultSchema

    async def aget(self, doc_number) -> BiblioResult:
        result = self.__schema__.load(await PublishedAsyncApi.biblio.get_biblio(doc_number))
        if len(result.documents) > 1:
            raise ValueError(f"More than one result found for {doc_number}!")
        return result.documents[0]


class ClaimsManager(Manager):
    __schema__ = ClaimsSchema

    async def aget(self, doc_number) -> Claims:
        return self.__schema__.load(await PublishedAsyncApi.fulltext.get_claims(doc_number))


class DescriptionManager(Manager):
    __schema__ = DescriptionSchema

    async def aget(self, doc_number) -> Description:
        return self.__schema__.load(await PublishedAsyncApi.fulltext.get_description(doc_number))


class ImageManager(Manager):
    __schema__ = ImagesSchema

    async def aget(self, doc_number) -> ImageDocument:
        return self.__schema__.load(await PublishedAsyncApi.images.get_images(doc_number))



import pytest

from .model import Patent
from .model import PatentBiblio
from .model import PublicSearch
from .model import PublishedApplication
from .model import PublishedApplicationBiblio


class TestPatents:
    def test_simple_lookup(self):
        app = PublicSearch.objects.get(patent_number="6103599")
        assert app.appl_id == "09089931"
        assert app.guid == "US-6103599-A"
        assert app.patent_title == "Planarizing technique for multilayered substrates"

    def test_tennis_patents(self):
        tennis_patents = Patent.objects.filter(title="tennis", assignee_name="wilson")
        assert len(tennis_patents) > 10

    def test_fetch_patent(self):
        pat = Patent.objects.get(6095661)
        assert pat.patent_title == "Method and apparatus for an L.E.D. flashlight"
        assert len(pat.abstract) == 1543

    def test_government_interest(self):
        pat = Patent.objects.get(11555621)
        assert pat.document.government_interest is not None

    def test_claims(self):
        pat_no = 6095661
        pat = Patent.objects.get(6095661)
        assert len(pat.claims) == 52
        assert pat.claims[0].number == 1
        assert len(pat.claims[0].text) == 1269
        assert len(pat.claims[0].limitations) == 5
        assert pat.claims[30].number == 31
        assert len(pat.claims[30].text) == 461
        assert len(pat.claims[30].limitations) == 4
        assert not pat.claims[30].dependent
        assert pat.claims[30].independent
        assert pat.claims[30].depends_on == list()
        assert pat.claims[31].dependent
        assert not pat.claims[31].independent
        assert pat.claims[31].depends_on == [
            31,
        ]

    def test_us8645300(self):
        pat_no = 8645300
        pat = Patent.objects.get(pat_no)
        assert pat.patent_title == "System and method for intent data processing"

    def test_search_classification(self):
        query = '"B60N2/5628".CPC. AND @APD>="20210101"<=20210107'
        results = Patent.objects.filter(query=query)
        assert len(results) >= 1
        assert results[0].publication_number == "11554698"

    def test_can_fetch_reissue_applications(self):
        pat_no = "RE43633"
        pat = Patent.objects.get(pat_no)
        assert (
            pat.patent_title
            == "System and method for linking streams of multimedia data to reference material for display"
        )
        assert pat.claims[0].text[:25] == "1. .[.A system for linkin"

    def test_can_get_field_of_search(self):
        pat_no = "9140110"
        pat = Patent.objects.get(pat_no)
        assert len(pat.field_of_search_us) == 6
        pat_no = "8832265"
        pat = Patent.objects.get(pat_no)
        assert len(pat.field_of_search_us) == 7

    def test_handles_disclaimers(self):
        """Some patents have notice markers on the issue date regarding statutory disclaimers"""
        pat = Patent.objects.get("5439055")
        assert pat.publication_date.isoformat() == "1995-08-08"

    def test_handles_claims_for_US6460631(self):
        pat = Patent.objects.get(6460631)
        for claim in pat.claims:
            pass
        assert True

    def test_handles_search_without_results(self):
        query = Patent.objects.filter(query='"379/56".CCLS. AND @APD>=19700101<=19950928')
        assert len(query) == 0

    def test_can_get_forward_references(self):
        pat = Patent.objects.get(6103599)
        forward_refs = pat.forward_citations
        assert len(forward_refs) >= 134

    def test_classifications_with_foreign_priority_data(self):
        pat = Patent.objects.get(7752445)
        assert len(pat.us_class_issued) == 6
        assert pat.foreign_priority[0].app_number == "2004-052835"

    def test_can_get_design_patents(self):
        pat = Patent.objects.get("D645062")
        assert pat.patent_title == "Gripping arm"
        assert pat.appl_id == "29380046"

    def test_search_that_has_single_patent(self):
        result = Patent.objects.filter(title="tennis", issue_date="2010-01-01->2010-02-27")
        assert len(result) == 1
        obj = result.first()
        assert obj.publication_number == "7658211"
        assert obj.patent_title == "Tennis ball recharging apparatus method"

    @pytest.mark.no_vcr
    def test_can_get_images(self, tmp_path):
        pat = Patent.objects.get("6103599")
        path = pat.download_images(path=tmp_path)
        assert path == tmp_path / "US-6103599-A.pdf"
        assert path.exists()


class TestPublishedApplicationFullText:
    def test_fetch_publication(self):
        pub_no = 20_160_009_839
        pub = PublishedApplication.objects.get(pub_no)
        assert (
            pub.patent_title == "POLYMER PRODUCTS AND MULTI-STAGE POLYMERIZATION PROCESSES FOR THE PRODUCTION THEREOF"
        )
        assert len(pub.abstract) == 651

    def test_publication_claims(self):
        pub_no = 20_160_009_839
        pub = PublishedApplication.objects.get(pub_no)
        assert len(pub.claims) == 25
        # Test Claim 1
        claim_1 = pub.claims[0]
        assert claim_1.number == 1
        assert len(claim_1.text) == 487
        assert len(claim_1.limitations) == 3
        assert claim_1.independent
        # Test Dependent Claim 2
        claim_2 = pub.claims[1]
        assert claim_2.number == 2
        assert claim_2.dependent

    def test_us_pub_20170370151(self):
        pub_no = 20_170_370_151
        pub = PublishedApplication.objects.get(pub_no)
        for i in range(6):
            assert pub.claims[i].text == f"{i+1}. (canceled)"
        assert "A system to control directional drilling in borehole drilling for" in pub.claims[6].text

    def test_search_classification(self):
        query = '"166/308.1".CCLS. AND @APD>=20150101<=20210101'
        results = PublishedApplicationBiblio.objects.filter(query=query)
        assert len(results) == 41
        assert len(list(results[:15])) == 15
        counter = 0
        for _ in results:
            counter += 1
        assert counter == 41

    def test_nonstandard_claim_format(self):
        obj = PublishedApplication.objects.get("20170260839")
        assert obj.claims[0].text[:39] == "1. A method of well ranging comprising:"

    @pytest.mark.no_vcr
    def test_can_get_images(self, tmp_path):
        pat = PublishedApplication.objects.get("20090150362")
        pat.download_images(path=tmp_path)
        assert (tmp_path / "US-20090150362-A1.pdf").exists()

    def test_links(self):
        pat = PatentBiblio.objects.get("6103599")
        assert pat.application.patent_number == "6103599"
        assert pat.global_dossier.app_num == "09089931"
        assert pat.assignments[0].id == "9218-376"
        assert pat.inpadoc.title == "Planarizing technique for multilayered substrates"

    def test_slices(self):
        pats = PatentBiblio.objects.filter(title="tennis", issue_date="2010-01-01->2010-12-31")
        assert pats[15] != pats[20]
        assert pats[0] != pats[1]

    def test_pages(self):
        pats = PatentBiblio.objects.filter(title="system").limit(525)
        old_p = None
        counter = 0
        for p in pats:
            counter += 1
            assert p != old_p
        assert counter == 525


class TestPatentsAsync:
    @pytest.mark.asyncio
    async def test_simple_lookup(self):
        app = await PublicSearch.objects.aget(patent_number="6103599")
        assert app.appl_id == "09089931"
        assert app.guid == "US-6103599-A"
        assert app.patent_title == "Planarizing technique for multilayered substrates"

    @pytest.mark.asyncio
    async def test_tennis_patents(self):
        tennis_patents = Patent.objects.filter(title="tennis", assignee_name="wilson")
        assert await tennis_patents.alen() > 10

    @pytest.mark.asyncio
    async def test_fetch_patent(self):
        pat = await Patent.objects.aget(6095661)
        assert pat.patent_title == "Method and apparatus for an L.E.D. flashlight"
        assert len(pat.abstract) == 1543

    @pytest.mark.asyncio
    async def test_government_interest(self):
        pat = await Patent.objects.aget(11555621)
        assert pat.document.government_interest is not None

    @pytest.mark.asyncio
    async def test_claims(self):
        pat_no = 6095661
        pat = await Patent.objects.aget(6095661)
        assert len(pat.claims) == 52
        assert pat.claims[0].number == 1
        assert len(pat.claims[0].text) == 1269
        assert len(pat.claims[0].limitations) == 5
        assert pat.claims[30].number == 31
        assert len(pat.claims[30].text) == 461
        assert len(pat.claims[30].limitations) == 4
        assert not pat.claims[30].dependent
        assert pat.claims[30].independent
        assert pat.claims[30].depends_on == list()
        assert pat.claims[31].dependent
        assert not pat.claims[31].independent
        assert pat.claims[31].depends_on == [
            31,
        ]

    @pytest.mark.asyncio
    async def test_us8645300(self):
        pat_no = 8645300
        pat = await Patent.objects.aget(pat_no)
        assert pat.patent_title == "System and method for intent data processing"

    @pytest.mark.asyncio
    async def test_search_classification(self):
        query = '"B60N2/5628".CPC. AND @APD>="20210101"<=20210107'
        results = Patent.objects.filter(query=query)
        assert await results.alen() >= 1
        app = await results.afirst()
        assert app.publication_number == "11554698"

    @pytest.mark.asyncio
    async def test_can_fetch_reissue_applications(self):
        pat_no = "RE43633"
        pat = await Patent.objects.aget(pat_no)
        assert (
            pat.patent_title
            == "System and method for linking streams of multimedia data to reference material for display"
        )
        assert pat.claims[0].text[:25] == "1. .[.A system for linkin"

    @pytest.mark.asyncio
    async def test_can_get_field_of_search(self):
        pat_no = "9140110"
        pat = await Patent.objects.aget(pat_no)
        assert len(pat.field_of_search_us) == 6
        pat_no = "8832265"
        pat = await Patent.objects.aget(pat_no)
        assert len(pat.field_of_search_us) == 7

    @pytest.mark.asyncio
    async def test_handles_disclaimers(self):
        """Some patents have notice markers on the issue date regarding statutory disclaimers"""
        pat = await Patent.objects.aget("5439055")
        assert pat.publication_date.isoformat() == "1995-08-08"

    @pytest.mark.asyncio
    async def test_handles_claims_for_US6460631(self):
        pat = await Patent.objects.aget(6460631)
        for claim in pat.claims:
            pass
        assert True

    @pytest.mark.asyncio
    async def test_handles_search_without_results(self):
        query = Patent.objects.filter(query='"379/56".CCLS. AND @APD>=19700101<=19950928')
        assert await query.alen() == 0

    @pytest.mark.asyncio
    async def test_can_get_forward_references(self):
        pat = await Patent.objects.aget(6103599)
        forward_refs = pat.forward_citations
        assert len(forward_refs) >= 134

    @pytest.mark.asyncio
    async def test_classifications_with_foreign_priority_data(self):
        pat = await Patent.objects.aget(7752445)
        assert len(pat.us_class_issued) == 6
        assert pat.foreign_priority[0].app_number == "2004-052835"

    @pytest.mark.asyncio
    async def test_can_get_design_patents(self):
        pat = await Patent.objects.aget("D645062")
        assert pat.patent_title == "Gripping arm"
        assert pat.appl_id == "29380046"

    @pytest.mark.asyncio
    async def test_search_that_has_single_patent(self):
        result = Patent.objects.filter(title="tennis", issue_date="2010-01-01->2010-02-27")
        assert await result.alen() == 1
        obj = await result.afirst()
        assert obj.publication_number == "7658211"
        assert obj.patent_title == "Tennis ball recharging apparatus method"

    @pytest.mark.no_vcr
    @pytest.mark.asyncio
    async def test_can_get_images(self, tmp_path):
        pat = await Patent.objects.aget("6103599")
        path = await pat.adownload_images(path=tmp_path)
        assert path == tmp_path / "US-6103599-A.pdf"
        assert path.exists()


class TestPublishedApplicationFullTextAsync:
    @pytest.mark.asyncio
    async def test_fetch_publication(self):
        pub_no = 20_160_009_839
        pub = await PublishedApplication.objects.aget(pub_no)
        assert (
            pub.patent_title == "POLYMER PRODUCTS AND MULTI-STAGE POLYMERIZATION PROCESSES FOR THE PRODUCTION THEREOF"
        )
        assert len(pub.abstract) == 651

    @pytest.mark.asyncio
    async def test_publication_claims(self):
        pub_no = 20_160_009_839
        pub = await PublishedApplication.objects.aget(pub_no)
        assert len(pub.claims) == 25
        # Test Claim 1
        claim_1 = pub.claims[0]
        assert claim_1.number == 1
        assert len(claim_1.text) == 487
        assert len(claim_1.limitations) == 3
        assert claim_1.independent
        # Test Dependent Claim 2
        claim_2 = pub.claims[1]
        assert claim_2.number == 2
        assert claim_2.dependent

    @pytest.mark.asyncio
    async def test_us_pub_20170370151(self):
        pub_no = 20_170_370_151
        pub = await PublishedApplication.objects.aget(pub_no)
        for i in range(6):
            assert pub.claims[i].text == f"{i+1}. (canceled)"
        assert "A system to control directional drilling in borehole drilling for" in pub.claims[6].text

    @pytest.mark.asyncio
    async def test_search_classification(self):
        query = '"166/308.1".CCLS. AND @APD>=20150101<=20210101'
        results = PublishedApplicationBiblio.objects.filter(query=query)
        assert await results.alen() == 41
        counter = 0
        async for _ in results:
            counter += 1
        assert counter == 41

    @pytest.mark.asyncio
    async def test_nonstandard_claim_format(self):
        obj = await PublishedApplication.objects.aget("20170260839")
        assert obj.claims[0].text[:39] == "1. A method of well ranging comprising:"

    @pytest.mark.no_vcr
    @pytest.mark.asyncio
    async def test_can_get_images(self, tmp_path):
        pat = await PublishedApplication.objects.aget("20090150362")
        await pat.adownload_images(path=tmp_path)
        assert (tmp_path / "US-20090150362-A1.pdf").exists()

    @pytest.mark.asyncio
    async def test_pages(self):
        pats = PatentBiblio.objects.filter(title="system").limit(525)
        old_p = None
        counter = 0
        async for p in pats:
            counter += 1
            assert p != old_p
        assert counter == 525



import csv
import datetime
from collections.abc import Sequence
from pathlib import Path

from dateutil.parser import parse as parse_dt
from dateutil.parser._parser import ParserError


class QueryException(Exception):
    pass


class QueryBuilder:
    def __init__(self):
        config_file = Path(__file__).parent / "query_config.csv"
        with config_file.open(encoding="utf-8-sig") as csvfile:
            reader = csv.DictReader(csvfile)
            config = list(reader)
        self.search_keywords = {r["keyword"]: r["query_field"] for r in config if r["query_field"]}
        self.order_by_keywords = {r["keyword"]: r["order_by_field"] for r in config if r["order_by_field"]}
        self.date_fields = [r["keyword"] for r in config if r["is_date"] == "X"]

    def convert_date(self, date):
        if isinstance(date, str):
            try:
                return parse_dt(date).strftime("%Y%m%d")
            except ParserError:
                raise QueryException(f"{date} is not a valid date!")
        elif isinstance(date, (datetime.date, datetime.datetime)):
            return date.strftime("%Y%m%d")
        else:
            raise QueryException(f"{date} is not a valid date!")

    def is_sequence(self, value):
        return isinstance(value, Sequence) and not isinstance(value, str)

    def query_value(self, key, value):
        field, *_ = key.split("__")
        if field in self.date_fields and "__" not in key:
            if isinstance(value, str) and "->" in value:
                start, end = value.split("->")
                return f"@{self.search_keywords[key]}>={self.convert_date(start)}<={self.convert_date(end)}"
            else:
                return f'@{self.search_keywords[key]}="{self.convert_date(value)}"'
        elif field in self.date_fields and "__" in key:
            field, modifier = key.split("__")
            if modifier == "range":
                if not self.is_sequence(value) and len(value) == 2:
                    raise ValueError(f"Input {value} is not a valid date range! Must be a tuple/list of length 2")
                # breakpoint()
                return f"@{self.search_keywords[field]}>={self.convert_date(value[0])}<={self.convert_date(value[1])}"
            elif modifier == "lt":
                return f'@{self.search_keywords[field]}<"{self.convert_date(value)}"'
            elif modifier == "lte":
                return f'@{self.search_keywords[field]}<="{self.convert_date(value)}"'
            elif modifier == "gt":
                return f'@{self.search_keywords[field]}>"{self.convert_date(value)}"'
            elif modifier == "gte":
                return f'@{self.search_keywords[field]}>="{self.convert_date(value)}"'
            else:
                raise ValueError(f"Modifier {modifier} is invalid! must be one of range, lt, lte, gt, gte")
        elif self.is_sequence(value) and len(value) > 1:
            value = " ".join(f'"{v}"' for v in value)
            return f"({value}).{self.search_keywords[key]}."
        elif self.is_sequence(value) and len(value) == 1:
            return f'"{value[0]}".{self.search_keywords[key]}.'
        else:
            return f'"{value}".{self.search_keywords[key]}.'

    def build_query(self, config):
        query_components = list()
        for key, value in config.filter.items():
            if key == "query":
                query_components.append(value)
                continue
            if key.split("__")[0] not in self.search_keywords:
                raise QueryException(f"{key} is not a valid search field!")
            else:
                query_components.append(self.query_value(key, value))
        default_operator = config.options.get("default_operator", "AND")
        return f" {default_operator} ".join(query_components)

    def build_order_by(self, config):
        if not config.order_by:
            return "date_publ desc"
        order_str = list()
        for value in config.order_by:
            if value.startswith("+"):
                order_str.append(f"{self.order_by_keywords[value[1:]]} asc")
            elif value.startswith("-"):
                order_str.append(f"{self.order_by_keywords[value[1:]]} desc")
            else:
                order_str.append(f"{self.order_by_keywords[value]} asc")
        return " ".join(order_str)



import pytest

from .api import PatentExaminationDataSystemApi


@pytest.mark.asyncio
async def test_can_get_app():
    result = await PatentExaminationDataSystemApi().create_query("applId:(16123456)")
    assert result.num_found == 1
    assert result.applications[0].appl_id == "16123456"


@pytest.mark.asyncio
async def test_can_search_by_customer_number():
    result = await PatentExaminationDataSystemApi().create_query("appCustNumber:(70155)")
    assert result.num_found > 10


@pytest.mark.asyncio
async def test_can_limit_by_rows():
    result = await PatentExaminationDataSystemApi().create_query("appCustNumber:(70155)", rows=5)
    assert len(result.applications) == 5



from .family.model import Family
from .legal.model import Legal
from .published.model import Images
from .published.model import Inpadoc

__api_name__ = "EPO Open Patent Services"
__all__ = ["Family", "Legal", "Images", "Inpadoc", "session"]



from patent_client.epo.ops.session import asession


class LegalAsyncApi:
    @classmethod
    async def get_legal(cls, doc_number, doc_type="publication", format="docdb"):
        url = f"http://ops.epo.org/3.2/rest-services/legal/{doc_type}/{format}/{doc_number}"
        response = await asession.get(url)
        response.raise_for_status()
        return response.text



from patent_client.epo.ops.util import Schema
from yankee.xml import fields as f


class InpadocSchema(Schema):
    family_id = f.Str("./@family-id")
    id_type = f.Str(".//epo:document-id/@document-id-type")
    country = f.Str(".//epo:document-id/epo:country")
    doc_number = f.Str(".//epo:document-id/epo:doc-number")
    kind = f.Str(".//epo:document-id/epo:kind")


class SearchSchema(Schema):
    query = f.Str(".//ops:query")
    num_results = f.Int(".//ops:biblio-search/@total-result-count")
    begin = f.Int(".//ops:range/@begin")
    end = f.Int(".//ops:range/@end")
    results = f.List(InpadocSchema, ".//ops:search-result/ops:publication-reference")



import json
from pathlib import Path

from patent_client.util.test import compare_dicts

from .schema import DocumentListSchema
from .schema import GlobalDossierSchema

test_dir = Path(__file__).parent / "test"


def test_app():
    data = (test_dir / "app.json").read_text()
    parser = GlobalDossierSchema()
    result = parser.load(data)
    expected_file = test_dir / "app_expected.json"
    # expected_file.write_text(result.to_json(indent=2))
    expected = json.loads(expected_file.read_text())
    result_json = json.loads(result.to_json())
    compare_dicts(expected, result_json)


def test_doc_list():
    data = (test_dir / "doc_list.json").read_text()
    parser = DocumentListSchema()
    result = parser.load(data)
    expected_file = test_dir / "doc_list_expected.json"
    expected_file.write_text(result.to_json(indent=2))
    expected = json.loads(expected_file.read_text())
    result_json = json.loads(result.to_json())
    compare_dicts(expected, result_json)



from typing import *

from patent_client.epo.ops.number_service.schema import DocumentIdSchema
from patent_client.epo.ops.util import Schema
from yankee.util import clean_whitespace
from yankee.xml import fields as f

from .national_codes import LegalCodes

code_db = LegalCodes()


def ip_type_formatter(string):
    types = {
        "PI": "Patent of Invention",
        "UM": "Utility Model",
    }
    return types.get(string, None)


def status_of_data_formatter(string):
    types = {"N": "New", "D": "Deleted", "O": "Optional", "C": "Backfile"}
    return types.get(string, None)


class DocumentNumberField(f.Combine):
    country = f.Str(".//ops:L001EP")
    doc_number = f.Str(".//ops:L003EP")
    kind_code = f.Str(".//ops:L004EP")

    def combine_func(self, obj):
        return f"{obj.country}{obj.doc_number}{obj.kind_code}"


class CorrespondingPatentField(f.Combine):
    doc_number = f.Str(".//ops:L503EP")
    country = f.Str(".//ops:L504EP")
    kind_code = f.Str(".//ops:L506EP")

    def combine_func(self, obj):
        if not obj.country and not obj.doc_number:
            return None
        return f"{obj.country}{obj.doc_number}{obj.get('kind_code', '')}"


class MetaDataSchema(Schema):
    status_of_data = f.Str(".//ops:L013", formatter=status_of_data_formatter)
    docdb_publication_number = f.Str(".//ops:L017EP")
    subscriber_exchange_date = f.Date(".//ops:L018EP")
    epo_created_date = f.Date(".//ops:L018EP")
    docdb_integer = f.Int(".//ops:L020EP")


class TextRecord(f.Combine):
    lines = f.List(f.Str(formatter=clean_whitespace), ".//ops:pre")

    def combine_func(self, obj):
        return "\n".join(obj.lines)


class LegalEventSchema(Schema):
    """
    Field descriptions are here:
    https://documents.epo.org/projects/babylon/eponot.nsf/0/EF223017D933B30AC1257B50005A042E/$File/14.11_User_Documentation_3.1_en.pdf

    """

    filing_or_publication = f.Str(".//ops:L002EP")
    document_number = DocumentNumberField()
    ip_type = f.Str(".//ops:L005EP", formatter=ip_type_formatter)
    metadata = MetaDataSchema()
    text_record = TextRecord()

    event_date = f.Date(".//ops:L007EP")
    event_code = f.Str(".//ops:L008EP")
    event_country = f.Str(".//ops:L501EP")
    country_code = f.Str(".//ops:L001EP")
    regional_event_code = f.Str(".//ops:L502EP")

    corresponding_patent = CorrespondingPatentField()
    corresponding_patent_publication_date = f.Date(".//ops:L505EP")

    designated_states = f.Str(".//ops:L507EP")
    extension_name = f.Str(".//ops:L508EP")
    new_owner_name = f.Str(".//ops:L509EP")
    free_text = f.Str(".//ops:L510EP")
    spc_number = f.Str(".//ops:L511EP")
    spc_filing_date = f.Str(".//ops:L512EP")
    expiry = f.Str(".//ops:L513EP")
    publication_language = f.Str(".//ops:L514EP")
    inventor_name = f.Str(".//ops:L515EP")
    ipc_class = f.Str(".//ops:L516EP")
    representative_name = f.Str(".//ops:L517EP")
    payment_date = f.Str(".//ops:L518EP")
    opponent_name = f.Str(".//ops:L519EP")
    year_of_fee_payment = f.Str(".//ops:L520EP")
    name_of_requester = f.Str(".//ops:L522EP")
    date_extension_granted = f.Str(".//ops:L523EP")
    extension_states = f.Str(".//ops:L524EP", formatter=lambda x: x.split(" "))
    effective_date = f.Date(".//ops:L525EP")
    date_of_withdrawal = f.Str(".//ops:L526EP")

    def deserialize(self, obj) -> "Dict":
        obj = super().deserialize(obj)
        if obj.event_code == "REG":
            code_data = code_db.get_code_data(obj.event_country, obj.regional_event_code)
            obj["event_code"] = obj.event_country + "." + obj.regional_event_code
        else:
            code_data = code_db.get_code_data(obj.country_code, obj.event_code)
            obj["event_code"] = obj.country_code + "." + obj.event_code
        obj["event_description"] = code_data["description"]
        return obj


class LegalSchema(Schema):
    publication_reference = DocumentIdSchema(".//ops:patent-family/ops:publication-reference")
    events = f.List(LegalEventSchema, ".//ops:legal")



import datetime
from dataclasses import dataclass
from dataclasses import field

from patent_client.util.asyncio_util import run_sync
from patent_client.util.base.model import Model
from patent_client.util.base.related import get_model
from patent_client.util.claims.parser import ClaimsParser
from yankee.data import ListCollection

from . import public_search_async_api

claim_parser = ClaimsParser()


@dataclass
class DocumentStructure(Model):
    number_of_claims: "Optional[int]" = None
    number_of_drawing_sheets: "Optional[int]" = None
    number_of_figures: "Optional[int]" = None
    page_count: "Optional[int]" = None
    front_page_end: "Optional[int]" = None
    front_page_start: "Optional[int]" = None
    bib_start: "Optional[int]" = None
    bib_end: "Optional[int]" = None
    abstract_start: "Optional[int]" = None
    abstract_end: "Optional[int]" = None
    drawings_start: "Optional[int]" = None
    drawings_end: "Optional[int]" = None
    description_start: "Optional[int]" = None
    description_end: "Optional[int]" = None
    specification_start: "Optional[int]" = None
    specification_end: "Optional[int]" = None
    claims_end: "Optional[int]" = None
    claims_start: "Optional[int]" = None
    amend_start: "Optional[int]" = None
    amend_end: "Optional[int]" = None
    cert_correction_end: "Optional[int]" = None
    cert_correction_start: "Optional[int]" = None
    cert_reexamination_end: "Optional[int]" = None
    cert_reexamination_start: "Optional[int]" = None
    ptab_start: "Optional[int]" = None
    ptab_end: "Optional[int]" = None
    search_report_start: "Optional[int]" = None
    search_report_end: "Optional[int]" = None
    supplemental_start: "Optional[int]" = None
    supplemental_end: "Optional[int]" = None


@dataclass
class PublicSearch(Model):
    __manager__ = "patent_client.uspto.public_search.manager.PublicSearchManager"
    guid: "Optional[str]" = None

    appl_id: "Optional[str]" = None
    app_filing_date: "Optional[datetime.date]" = None
    related_appl_filing_date: "ListCollection" = field(default_factory=ListCollection)
    publication_number: "Optional[str]" = None
    kind_code: "Optional[str]" = None
    publication_date: "Optional[datetime.date]" = None
    patent_title: "Optional[str]" = None

    inventors_short: "Optional[str]" = None
    applicant_name: "ListCollection" = field(default_factory=ListCollection)
    assignee_name: "ListCollection" = field(default_factory=ListCollection)
    government_interest: "ListCollection" = field(default_factory=ListCollection)
    primary_examiner: "Optional[str]" = None
    assistant_examiner: "ListCollection" = field(default_factory=ListCollection)

    main_classification_code: "Optional[str]" = None
    cpc_additional: "Optional[str]" = None
    cpc_inventive: "Optional[str]" = None
    ipc_code: "Optional[str]" = None
    uspc_full_classification: "Optional[str]" = None

    image_file_name: "Optional[str]" = None
    image_location: "Optional[str]" = None
    document_structure: "Optional[DocumentStructure]" = None

    type: "Optional[str]" = None
    database_name: "Optional[str]" = None
    composite_id: "Optional[str]" = None
    document_id: "Optional[str]" = None
    document_size: "Optional[int]" = None
    family_identifier_cur: "Optional[int]" = None
    language_indicator: "Optional[str]" = None

    score: "Optional[float]" = None

    def __repr__(self):
        return f"PublicationBiblio(publication_number={self.publication_number}, publication_date={self.publication_date.isoformat()}, patent_title={self.patent_title})"

    @property
    def document(self):
        return get_model("patent_client.uspto.public_search.model.PatentDocument").objects.get(guid=self.guid)

    @property
    def forward_citations(self):
        return get_model("patent_client.uspto.public_search.model.PublicSearch").objects.filter(
            us_reference=self.publication_number
        )

    @property
    def application(self):
        return get_model("patent_client.uspto.peds.model.USApplication").objects.get(self.appl_id)

    @property
    def global_dossier(self):
        return get_model("patent_client.uspto.global_dossier.model.GlobalDossierApplication").objects.get(self.appl_id)

    @property
    def assignments(self):
        return get_model("patent_client.uspto.assignment.model.Assignment").objects.filter(appl_id=self.appl_id)

    @property
    def inpadoc(self):
        return get_model("patent_client.epo.ops.published.model.Inpadoc").objects.get("US" + self.publication_number)

    def download_images(self, path="."):
        return run_sync(public_search_async_api.download_image(self, path))

    async def adownload_images(self, path="."):
        return await public_search_async_api.download_image(self, path)


@dataclass
class Document(Model):
    abstract: "Optional[str]" = None
    government_interest: "Optional[str]" = None
    background: "Optional[str]" = None
    description: "Optional[str]" = None
    brief: "Optional[str]" = None
    claim_statement: "Optional[str]" = None
    claims: "Optional[str]" = None


@dataclass
class UsReference(Model):
    publication_number: "Optional[str]" = None
    pub_month: "Optional[datetime.date]" = None
    patentee_name: "Optional[str]" = None
    cited_by_examiner: "Optional[bool]" = None


@dataclass
class ForeignReference(Model):
    citation_classification: "Optional[str]" = None
    citation_cpc: "Optional[str]" = None
    country_code: "Optional[str]" = None
    patent_number: "Optional[str]" = None
    pub_month: "Optional[datetime.date]" = None
    cited_by_examiner: "Optional[bool]" = None


@dataclass
class NplReference(Model):
    citation: "Optional[str]" = None
    cited_by_examiner: "Optional[bool]" = None


@dataclass
class RelatedApplication(Model):
    child_patent_country: "Optional[str]" = None
    child_patent_number: "Optional[str]" = None
    country_code: "Optional[str]" = None
    filing_date: "Optional[datetime.date]" = None
    number: "Optional[str]" = None
    parent_status_code: "Optional[str]" = None
    patent_issue_date: "Optional[datetime.date]" = None
    patent_number: "Optional[str]" = None


@dataclass
class Inventor(Model):
    name: "Optional[str]" = None
    city: "Optional[str]" = None
    country: "Optional[str]" = None
    postal_code: "Optional[str]" = None
    state: "Optional[str]" = None


@dataclass
class Applicant(Model):
    city: "Optional[str]" = None
    country: "Optional[str]" = None
    # group: "Optional[str]" = None
    name: "Optional[str]" = None
    state: "Optional[str]" = None
    zip_code: "Optional[str]" = None
    authority_type: "Optional[str]" = None


@dataclass
class Assignee(Model):
    city: "Optional[str]" = None
    country: "Optional[str]" = None
    name: "Optional[str]" = None
    postal_code: "Optional[str]" = None
    state: "Optional[str]" = None
    type_code: "Optional[str]" = None


@dataclass
class CpcCode(Model):
    cpc_class: "Optional[str]" = None
    cpc_subclass: "Optional[str]" = None
    version: "Optional[datetime.date]" = None


@dataclass
class IntlCode(Model):
    intl_class: "Optional[str]" = None
    intl_subclass: "Optional[str]" = None
    version: "Optional[datetime.date]" = None


@dataclass
class ForeignPriorityApplication(Model):
    country: "Optional[str]" = None
    app_filing_date: "Optional[datetime.date]" = None
    app_number: "Optional[str]" = None


@dataclass
class PublicSearchDocument(Model):
    guid: "Optional[str]" = None
    publication_number: "Optional[str]" = None
    publication_date: "Optional[datetime.date]" = None

    appl_id: "Optional[str]" = None
    patent_title: "Optional[str]" = None
    app_filing_date: "Optional[datetime.date]" = None
    application_type: "Optional[str]" = None
    family_identifier_cur: "Optional[int]" = None
    related_apps: "ListCollection[RelatedApplication]" = field(default_factory=ListCollection)
    foreign_priority: "ListCollection[ForeignPriorityApplication]" = field(default_factory=ListCollection)
    type: "Optional[str]" = None

    # Parties
    inventors: "ListCollection[Inventor]" = field(default_factory=ListCollection)
    inventors_short: "Optional[str]" = None
    applicants: "ListCollection[Applicant]" = field(default_factory=ListCollection)
    assignees: "ListCollection[Assignee]" = field(default_factory=ListCollection)

    group_art_unit: "Optional[str]" = None
    primary_examiner: "Optional[str]" = None
    assistant_examiner: "ListCollection" = field(default_factory=ListCollection)
    legal_firm_name: "ListCollection" = field(default_factory=ListCollection)
    attorney_name: "ListCollection" = field(default_factory=ListCollection)

    # Text Data
    document: "Optional[Document]" = None
    document_structure: "Optional[DocumentStructure]" = None

    # Image Data
    image_file_name: "Optional[str]" = None
    image_location: "Optional[str]" = None

    # Metadata
    composite_id: "Optional[str]" = None
    database_name: "Optional[str]" = None
    derwent_week_int: "Optional[int]" = None

    # References Cited
    us_references: "ListCollection[UsReference]" = field(default_factory=ListCollection)
    foreign_references: "ListCollection[ForeignReference]" = field(default_factory=ListCollection)
    npl_references: "ListCollection[NplReference]" = field(default_factory=ListCollection)

    # Classifications
    cpc_inventive: "ListCollection" = field(default_factory=ListCollection)
    cpc_additional: "ListCollection" = field(default_factory=ListCollection)

    intl_class_issued: "ListCollection[str]" = field(default_factory=ListCollection)
    intl_class_current_primary: "ListCollection[IntlClass]" = field(default_factory=ListCollection)
    intl_class_currrent_secondary: "ListCollection" = field(default_factory=ListCollection)

    us_class_current: "ListCollection[str]" = field(default_factory=ListCollection)
    us_class_issued: "ListCollection" = field(default_factory=ListCollection)

    field_of_search_us: "ListCollection" = field(default_factory=ListCollection)
    field_of_search_cpc: "ListCollection" = field(default_factory=ListCollection)

    def __repr__(self):
        return f"Publication(publication_number={self.publication_number}, publication_date={self.publication_date.isoformat()}, patent_title={self.patent_title})"

    @property
    def abstract(self):
        return self.document.abstract

    @property
    def description(self):
        sections = (
            "government_interest",
            "background",
            "brief",
            "description",
        )
        section_text = tuple(getattr(self.document, s) for s in sections)
        return "\n\n".join(s for s in section_text if s)

    @property
    def claims_text(self):
        return self.document.claims

    @property
    def claims(self):
        return claim_parser.parse(self.document.claims)

    @property
    def forward_citations(self):
        return get_model("patent_client.uspto.public_search.model.PublicSearch").objects.filter(
            us_reference=self.publication_number
        )

    def download_images(self, path="."):
        return run_sync(self.adownload_images(path))

    async def adownload_images(self, path="."):
        return await public_search_async_api.download_image(self, path)

    @property
    def application(self):
        return get_model("patent_client.uspto.peds.model.USApplication").objects.get(self.appl_id)

    @property
    def global_dossier(self):
        return get_model("patent_client.uspto.global_dossier.model.GlobalDossierApplication").objects.get(self.appl_id)

    @property
    def assignments(self):
        return get_model("patent_client.uspto.assignment.model.Assignment").objects.filter(appl_id=self.appl_id)

    @property
    def inpadoc(self):
        return get_model("patent_client.epo.ops.published.model.Inpadoc").objects.get("US" + self.publication_number)


@dataclass
class PatentBiblio(PublicSearch):
    __manager__ = "patent_client.uspto.public_search.manager.PatentBiblioManager"


@dataclass
class Patent(PublicSearchDocument):
    __manager__ = "patent_client.uspto.public_search.manager.PatentManager"


@dataclass
class PublishedApplicationBiblio(PublicSearch):
    __manager__ = "patent_client.uspto.public_search.manager.PublishedApplicationBiblioManager"


@dataclass
class PublishedApplication(PublicSearchDocument):
    __manager__ = "patent_client.uspto.public_search.manager.PublishedApplicationManager"



import datetime
from pathlib import Path

import requests_cache
from patent_client import SETTINGS
from patent_client.version import __version__
from requests.adapters import HTTPAdapter
from urllib3.util.retry import Retry

max_age = SETTINGS.CACHE.MAX_AGE

if isinstance(max_age, str) and "day" in max_age.lower():
    max_age = int(max_age.split()[0])


class PatentClientSession(requests_cache.CachedSession):
    def __init__(self):
        super().__init__(
            Path(SETTINGS.DEFAULT.BASE_DIR).expanduser() / SETTINGS.CACHE.PATH,
            expire_after=datetime.timedelta(days=max_age),
            backend="sqlite",
            allowable_methods=("GET", "POST"),
            ignored_parameters=[
                "Authorization",
            ],
        )
        self.headers[
            "User-Agent"
        ] = f"Mozilla/5.0 Python Patent Clientbot/{__version__} (parkerhancock@users.noreply.github.com)"
        # Install a default retry on the session using urrlib3
        retry = Retry(total=5, backoff_factor=0.2)
        self.mount("https://", HTTPAdapter(max_retries=retry))
        self.mount("http://", HTTPAdapter(max_retries=retry))



import json
from pathlib import Path

from patent_client.util.test import compare_dicts

from .schema import PedsPageSchema

test_dir = Path(__file__).parent / "test"


def test_us_app_12721698():
    data = (test_dir / "app_12721698.json").read_text()
    parser = PedsPageSchema()
    result = parser.load(data)
    expected_file = test_dir / "expected.json"
    # expected_file.write_text(result.to_json(indent=2))
    expected = json.loads(expected_file.read_text())
    compare_dicts(expected, json.loads(result.to_json()))



import json
from pathlib import Path

from patent_client.util.test import compare_lists

from .parser import ClaimsParser

example_dir = Path(__file__).parent / "examples"


def test_multiple_dependent_claims():
    file = example_dir / "multiple_dependent.txt"
    text = file.read_text()
    result = ClaimsParser().parse(text)
    expected_file = example_dir / "multiple_dependent.json"
    # expected_file.write_text(result.to_json(indent=2))
    expected = json.loads(expected_file.read_text())
    compare_lists(json.loads(result.to_json()), expected)


def test_published_claims():
    file = example_dir / "published_claims.txt"
    text = file.read_text()
    result = ClaimsParser().parse(text)
    expected_file = example_dir / "published_claims.json"
    # expected_file.write_text(result.to_json(indent=2))
    expected = json.loads(expected_file.read_text())
    compare_lists(json.loads(result.to_json()), expected)



from __future__ import annotations

import datetime
from dataclasses import dataclass
from dataclasses import field
from pathlib import Path
from typing import Iterable
from typing import List
from typing import Optional

from dateutil.relativedelta import relativedelta
from patent_client import session
from patent_client.util import Model
from patent_client.util.base.related import get_model
from yankee.data import ListCollection


@dataclass
class USApplication(Model):
    """A U.S. Patent Application retrieved from the Patent Examination Data
    System (PEDS)
    """

    __manager__ = "patent_client.uspto.peds.manager.USApplicationManager"

    appl_id: str = field(compare=True)
    """The application number. U.S. Applications are digits only.
    PCT numbers are in the format PCT/CCYY/#####"""

    app_filing_date: Optional[datetime.date] = field(default=None, repr=False)
    """The filing date or 371(c) date"""

    patent_title: "Optional[str]" = None
    """Title of the invention"""

    app_status: "Optional[str]" = field(default=None, repr=True)
    """Status of the Application"""

    app_status_date: Optional[datetime.date] = field(default=None, repr=False)
    """The date of the applicable status (the app_status attribute)"""

    app_early_pub_number: "Optional[str]" = field(default=None, repr=False)
    """The published patent application number in the format USYYYY#######A1
    Note: this does not include subsequent or corrected publications, or publications
    of PCT applications."""

    app_early_pub_date: Optional[datetime.date] = field(default=None, repr=False)
    """The publication date of the publication mentioned in app_early_pub_number"""

    patent_number: "Optional[str]" = field(default=None, repr=False)
    """The issued patent number, if any. Digits only"""

    patent_issue_date: Optional[datetime.date] = field(default=None, repr=False)
    """The date the patent issued"""

    wipo_early_pub_number: "Optional[str]" = field(default=None, repr=False)
    """If the application was published by WIPO (i.e. a PCT application),
    the publication number is here. Format is YYYY######"""

    wipo_early_pub_date: Optional[datetime.date] = field(default=None, repr=False)
    """Publication date by WIPO"""

    corr_addr_cust_no: "Optional[str]" = field(default=None, repr=False)
    """The customer number for the correspondent named for this application"""

    app_cust_number: "Optional[str]" = field(default=None, repr=False)
    """The customer number for the firm with authority to act in the application"""

    app_attr_dock_number: "Optional[str]" = field(default=None, repr=False)
    """The docket number assigned to the application by the handling firm"""

    app_location: "Optional[str]" = field(default=None, repr=False)
    """The physical location of the application file"""

    first_inventor_file: "Optional[str]" = field(default=None, repr=False)
    """Whether the application is examined under the America Invents Act"""

    app_type: "Optional[str]" = field(default=None, repr=False)
    """The subject matter of the application (i.e. Utility / Design / Plant)"""

    app_entity_status: "Optional[str]" = field(default=None, repr=False)
    """The entity type of the owner (UNDISCOUNTED / SMALL / MICRO)"""

    app_confr_number: "Optional[str]" = field(default=None, repr=False)
    """The application's confirmation number"""

    app_cls_sub_cls: "Optional[str]" = field(default=None, repr=False)
    """The U.S. classification assigned to the application"""

    app_grp_art_number: "Optional[str]" = field(default=None, repr=False)
    """The Group Art Unit to which the application is assigned"""

    app_exam_name: "Optional[str]" = field(default=None, repr=False)
    """The name of the examiner handling the application"""

    # Nested Objects
    pta_pte_summary: Optional[PtaPteSummary] = field(default=None, repr=False)
    """A related object containing the PTA/PTE analysis"""

    correspondent: Optional[Correspondent] = field(default=None, repr=False)
    """Correspondent for the Application"""

    # List Properties
    inventors: ListCollection[Inventor] = field(default=None, repr=False)
    """List of named inventors"""

    applicants: ListCollection[Applicant] = field(default_factory=list, repr=False)
    """List of named applicants"""

    attorneys: ListCollection[Attorney] = field(default_factory=list, repr=False)
    """List of attorneys authorized to act in this application, associated with
    the indicated customer number"""

    transactions: ListCollection[Transaction] = field(default_factory=list, repr=False)
    """List of transactions relating to this application. Identical to the "Transactions" tab on
    Patent Center or Private PAIR"""

    child_continuity: ListCollection[Relationship] = field(default_factory=ListCollection, repr=False)
    """List of related Applications which claim priority to this application. Note that
    this does not include continuity type (e.g. CON/CIP/DIV)"""

    parent_continuity: ListCollection[Relationship] = field(default_factory=ListCollection, repr=False)
    """List of related Applications that this application claims priority to, including
    continuity type. Does not include foreign priority claims"""

    foreign_priority: ListCollection[ForeignPriority] = field(default_factory=list, repr=False)
    """List of foreign patent applications to which this application claims priority"""

    pta_pte_tran_history: ListCollection[PtaPteHistory] = field(default_factory=list, repr=False)
    """List of transactions relevant to calculating a Patent Term Extension or Adjustment"""

    assignments: ListCollection[Assignment] = field(default_factory=ListCollection, repr=False)
    """List of Assignments that include this application"""

    def __hash__(self):
        return hash(self.appl_id)

    # Dynamic Properties

    @property
    def patent_center_link(self) -> str:
        return f"https://patentcenter.uspto.gov/applications/{self.appl_id}"

    @property
    def google_patents_link(self) -> str:
        if self.patent_number is not None:
            return f"https://patents.google.com/patent/US{self.patent_number}/en"
        elif self.app_early_pub_number is not None:
            return f"https://patents.google.com/patent/{self.app_early_pub_number}/en"
        else:
            return None

    @property
    def continuity(self) -> ListCollection[USApplication]:
        """Returns a complete set of parents, self, and children"""
        return ListCollection(
            [
                self.child_continuity.values_list("child", flat=True),
                [
                    self,
                ],
                self.parent_continuity.values_list("parent", flat=True),
            ]
        )

    @property
    def kind(self) -> str:
        """Differentiates provisional, PCT, and nonprovisional applications"""
        if "PCT" in self.appl_id:
            return "PCT"
        if self.appl_id[0] == "6":
            return "Provisional"
        return "Nonprovisional"

    @property
    def publication_number(self):
        return self.app_early_pub_number[2:-2]

    @property
    def priority_date(self) -> datetime.date:
        """Attempts to return the priority date of the application, calculated as
        the earliest application filing date among the application's parents, or
        its own filing date if it has no parents. Does not include foreign priority
        """
        if not self.parent_continuity:
            return self.app_filing_date
        else:
            return sorted(p.parent_app_filing_date for p in self.parent_continuity)[0]

    @property
    def expiration(self) -> Optional[Expiration]:
        """Calculates expiration data from which the expiration date can be calculated. See
        help information for the resulting Expiration model.
        """
        if "PCT" in self.appl_id:
            raise NotImplementedError("Expiration date not supported for PCT Applications")
        if not self.patent_number:
            return None
        expiration_data = dict()
        term_parents = [
            p
            for p in self.parent_continuity
            if p.relationship not in ["Claims Priority from Provisional Application", "is a Reissue of"]
        ]
        if term_parents:
            term_parent = sorted(term_parents, key=lambda x: x.parent_app_filing_date)[0]
            relationship = term_parent.relationship
            parent_filing_date = term_parent.parent_app_filing_date
            parent_appl_id = term_parent.parent_appl_id
        else:
            relationship = "self"
            parent_appl_id = self.appl_id
            parent_filing_date = self.app_filing_date

        expiration_data["parent_appl_id"] = parent_appl_id
        expiration_data["parent_app_filing_date"] = parent_filing_date
        expiration_data["parent_relationship"] = relationship
        expiration_data["initial_term"] = parent_filing_date + relativedelta(years=20)  # type: ignore
        expiration_data["pta_or_pte"] = self.pta_pte_summary.total_days or 0  # type: ignore
        expiration_data["extended_term"] = expiration_data["initial_term"] + relativedelta(
            days=expiration_data["pta_or_pte"]
        )  # type: ignore

        transactions = self.transactions
        try:
            disclaimer = next(t for t in transactions if t.code == "DIST")
            expiration_data["terminal_disclaimer_filed"] = True
        except StopIteration:
            expiration_data["terminal_disclaimer_filed"] = False

        return Expiration(**expiration_data)  # type: ignore

    # Related objects
    @property
    def documents(self) -> "Iterable[patent_client.uspto.peds.model.Document]":
        """File History Documents from PEDS CMS"""
        return get_model("patent_client.uspto.peds.model.Document").objects.filter(appl_id=self.appl_id)

    @property
    def related_assignments(
        self,
    ) -> "Iterable[patent_client.uspto.assignment.model.Assignment]":
        """Related Assignments from the Assignments API"""
        return get_model("patent_client.uspto.assignment.model.Assignment").objects.filter(appl_id=self.appl_id)

    @property
    def ptab_proceedings(
        self,
    ) -> "Iterable[patent_client.uspto.ptab.model.PtabProceeding]":
        """Related PtabProceedings for this application"""
        return get_model("patent_client.uspto.ptab.model.PtabProceeding").objects.filter(appl_id=self.appl_id)

    @property
    def patent(self) -> "Optional[patent_client.uspto.fulltext.patent.model.Patent]":
        """Fulltext version of the patent - If Available"""
        return get_model("patent_client.uspto.fulltext.patent.model.Patent").objects.get(
            publication_number=self.patent_number
        )

    @property
    def publication(
        self,
    ) -> "Optional[patent_client.uspto.fulltext.published_application.model.PublishedApplication]":
        """Fulltext version of the Publication - If Available"""
        return get_model(
            "patent_client.uspto.fulltext.published_application.model.PublishedApplication",
            publication_number=self.publication_number,
        )

    @property
    def inpadoc_patent(
        self,
    ) -> "Optional[patent_client.epo.ops.published.model.InpadocBiblio]":
        """Fulltext version of the patent - If Available"""
        return get_model(
            "patent_client.epo.ops.published.model.InpadocBiblio",
            publication_number=f"US{self.patent_number}",
        )

    @property
    def inpadoc_publication(
        self,
    ) -> "Optional[patent_client.epo.ops.published.model.InpadocBiblio]":
        """Fulltext version of the patent - If Available"""
        return get_model(
            "patent_client.epo.ops.published.model.InpadocBiblio",
            publication_number=self.app_early_pub_number,
        )


@dataclass
class Expiration(Model):
    """Model for patent application expiration data.
    NOTE: THIS IS NOT LEGAL ADVICE. See MPEP Sec. 2701 for a detailed explanation of calculating patent term

    This model provides a first-cut estimate of patent term by calulating four things:

    1. The earliest-filed non-provisional patent application (EFNP) from which term should be calculated (parent).
    2. 20 years from the filing date of the EFNP (initial term)
    3. Extensions from Patent Term Extentions (PTE) or Patent Term Adjustments (PTA) (extended_term)
    4. The presence or absence of a terminal disclaimer in the transaction history

    """

    parent_appl_id: str
    """Application number for the earliest-filed nonprovisional application related to this application, or self"""
    parent_app_filing_date: datetime.date
    """Filing date of the earliest-filed nonprovisional application related to this application, or self"""
    parent_relationship: str
    """Relationship of the earliest-filed nonprovisional application. Can be self"""
    initial_term: datetime.date
    """Patent term calculated as 20 years from earliest-field non-provisional (no adjustments)"""
    pta_or_pte: float
    """Days of extended patent term from a Patent Term Extension (PTE) or Patent Term Adjustment (PTA)"""
    extended_term: datetime.date
    """Patent term as extended by any applicable Patent Term Extension or Patent Term Adjustment"""
    terminal_disclaimer_filed: bool
    """Presence or absence of a terminal disclaimer in the transaction history of the application"""


@dataclass
class Relationship(Model):
    parent_appl_id: str
    """The application number of the parent application"""
    child_appl_id: str
    """The application number of the child application"""
    relationship: "Optional[str]"
    """The relationship of the parent to child. This information
    is only populated on Relationships associated with the child application"""
    child_app_filing_date: "Optional[datetime.date]" = None
    """The filing date of the child application"""
    parent_app_filing_date: "Optional[datetime.date]" = None
    """The filing date of the parent application"""
    parent_app_status: "Optional[str]" = None
    """The status of the parent application - which may not be indicated"""
    child_app_status: "Optional[str]" = None
    """The status of the chidl application - which may not be indicated"""

    @property
    def parent(self) -> "patent_client.uspto.peds.USApplication":
        """Link to the parent application"""
        return get_model("patent_client.uspto.peds.USApplication").objects.get(appl_id=self.parent_appl_id)

    @property
    def child(self) -> "patent_client.uspto.peds.USApplication":
        """Link to the child application"""
        return get_model("patent_client.uspto.peds.USApplication").objects.get(appl_id=self.child_appl_id)

    def __eq__(self, other):
        return (
            self.parent_appl_id == other.parent_appl_id
            and self.child_appl_id == other.child_appl_id
            and self.relationship == other.relationship
        )

    def __hash__(self):
        return hash((self.parent_appl_id, self.child_appl_id, self.relationship))


@dataclass
class ForeignPriority(Model):
    priority_claim: str = None
    """The application number of the foreign priority application"""
    country_name: str = None
    """The country in which the foreign priorty application was filed"""
    filing_date: datetime.date = None
    """The filing date of the foreign priority application"""


@dataclass
class PtaPteHistory(Model):
    number: float
    date: datetime.date
    description: str
    pto_days: "Optional[float]" = None
    applicant_days: "Optional[float]" = None
    start: "Optional[float]" = None


@dataclass
class PtaPteSummary(Model):
    a_delay: "Optional[int]" = None
    b_delay: "Optional[int]" = None
    c_delay: "Optional[int]" = None
    overlap_delay: "Optional[int]" = None
    pto_delay: "Optional[int]" = None
    applicant_delay: "Optional[int]" = None
    pto_adjustments: "Optional[int]" = None
    total_days: "Optional[int]" = None
    kind: "Optional[str]" = None


@dataclass
class Transaction(Model):
    date: datetime.date
    code: str
    description: str


@dataclass
class Correspondent(Model):
    name: str
    address: "Optional[str]" = None
    cust_no: "Optional[str]" = None


@dataclass
class Attorney(Model):
    name: str
    phone_num: str
    reg_status: "Optional[str]" = None
    registration_no: Optional[int] = None


@dataclass
class Applicant(Model):
    name: "Optional[str]" = None
    cust_no: "Optional[str]" = None
    address: "Optional[str]" = None
    rank_no: Optional[int] = None


@dataclass
class Inventor(Model):
    name: "Optional[str]" = None
    address: "Optional[str]" = None
    rank_no: Optional[int] = None


class PedsError(Exception):
    pass


@dataclass
class Document(Model):
    __manager__ = "patent_client.uspto.peds.manager.DocumentManager"
    base_url = "https://ped.uspto.gov/api/queries/cms/public/"
    access_level_category: str
    appl_id: str
    category: str
    code: str
    description: str
    identifier: str
    mail_room_date: datetime.date
    page_count: int
    url: "Optional[str]" = None

    @property
    def application(self) -> "patent_client.uspto.peds.model.USApplication":
        return get_model("patent_client.uspto.peds.model.USApplication").objects.get(appl_id=self.appl_id)

    def __repr__(self):
        return f"Document(appl_id={self.appl_id}, mail_room_date={self.mail_room_date}, description={self.description})"

    def download(self, path=".", include_appl_id=True):
        if str(path)[-4:].lower() == ".pdf":
            # If we've been given a specific filename, use it
            out_file = Path(path)
        elif include_appl_id:
            out_file = (
                Path(path) / f"{self.appl_id} - {self.mail_room_date} - {self.code} - {self.description[:40]}.pdf"
            )
        else:
            out_file = Path(path) / f"{self.mail_room_date} - {self.code} - {self.description[:40]}.pdf"

        with session.get(self.base_url + self.url, stream=True) as r:
            if r.status_code == 403:
                raise PedsError("File history document downloading is broken. This is a USPTO problem :(")
            r.raise_for_status()
            with out_file.open("wb") as f:
                for chunk in r.iter_content(chunk_size=8192):
                    f.write(chunk)
        return out_file


@dataclass
class Assignee(Model):
    name: str = None
    address: str = None


@dataclass
class Assignor(Model):
    name: str = None
    exec_date: "datetime.date" = None


@dataclass
class Assignment(Model):
    id: str
    correspondent: str = None
    correspondent_address: str = None
    mail_date: "datetime.date" = None
    received_date: "datetime.date" = None
    recorded_date: "datetime.date" = None
    pages: int = None
    conveyance_text: str = None
    sequence_number: int = None
    assignors: "ListCollection[Assignor]" = field(default_factory=ListCollection)
    assignees: "ListCollection[Assignee]" = field(default_factory=ListCollection)


@dataclass
class PedsPage(Model):
    index_last_updated: datetime.date
    num_found: int
    applications: "List[USApplication]" = field(default_factory=list)



import pytest

from .api import aconvert_number


class TestNumberServiceAsyncApi:
    @pytest.mark.asyncio
    async def test_docdb_to_epodoc(self):
        actual = await aconvert_number("MD.20050130.A.20050130", "application", "docdb", "epodoc")
        assert not actual.messages
        assert actual.input_doc.id_type == "docdb"
        assert actual.output_doc.id_type == "epodoc"

    @pytest.mark.asyncio
    async def test_original_to_docdb(self):
        actual = await aconvert_number("JP.(2006-147056).A.20060526", "application", "original", "docdb")
        assert not actual.messages
        assert actual.input_doc.id_type == "original"
        assert actual.output_doc.id_type == "docdb"

    @pytest.mark.asyncio
    async def test_docdb_to_original(self):
        actual = await aconvert_number("JP.2006147056.A.20060526", "application", "docdb", "original")
        assert not actual.messages
        assert actual.input_doc.id_type == "docdb"
        assert actual.output_doc.id_type == "original"

    @pytest.mark.asyncio
    async def test_original_to_epodoc(self):
        actual = await aconvert_number("US.(08/921,321).A.19970829", "application", "original", "epodoc")
        assert not actual.messages
        assert actual.input_doc.id_type == "original"
        assert actual.output_doc.id_type == "epodoc"

    @pytest.mark.asyncio
    async def test_original_to_docdb_pct(self):
        actual = await aconvert_number("PCT/GB02/04635.20021011", "application", "original", "docdb")
        assert actual.input_doc.id_type == "original"
        assert actual.output_doc.id_type == "docdb"



import json
from pathlib import Path

__api_name__ = "PTAB API v2"
base_dir = Path(__file__).parent
schema_path = base_dir / "ptabApiV2.json"
schema_doc = json.loads(schema_path.read_text())

from .model import PtabDecision, PtabDocument, PtabProceeding  # noqa: F401,E402



from patent_client.util import Manager

from .api import FamilyAsyncApi
from .schema import FamilySchema


class FamilyManager(Manager):
    __schema__ = FamilySchema

    async def aget(self, doc_number):
        return self.__schema__.load(await FamilyAsyncApi.get_family(doc_number, doc_type="publication", format="docdb"))



from yankee.json import fields as f
from yankee.json import Schema


class AdditionalRespondentSchema(Schema):
    application_number_text = f.Str()
    inventor_name = f.Str()
    patent_number = f.Str()
    party_name = f.Str()


class PtabProceedingSchema(Schema):
    # Proceeding Metadata
    last_modified_date = f.DateTime()
    last_modified_user_id = f.DateTime()
    institution_decision_date = f.Date()
    proceeding_filing_date = f.Date()
    accorded_filing_date = f.Date()
    proceeding_status_category = f.Str()
    proceeding_number = f.Str()
    proceeding_last_modified_date = f.Date()
    proceeding_type_category = f.Str()
    subproceeding_type_category = f.Str()
    decision_date = f.Date()
    docket_notice_mail_date = f.Date()
    declaration_date = f.Date()
    style_name_text = f.Str()

    # Respondent Information
    respondent_technology_center_number = f.Str()
    respondent_patent_owner_name = f.Str()
    respondent_party_name = f.Str()
    respondent_group_art_unit_number = f.Str()
    respondent_inventor_name = f.Str()
    respondent_counsel_name = f.Str()
    respondent_grant_date = f.Date()
    respondent_patent_number = f.Str()
    respondent_application_number_text = f.Str()
    respondent_publication_number = f.Str()
    respondent_publication_date = f.Date()

    # Petitioner Information
    petitioner_technology_center_number = f.Str()
    petitioner_patent_owner_name = f.Str()
    petitioner_party_name = f.Str()
    petitioner_group_art_unit_number = f.Str()
    petitioner_inventor_name = f.Str()
    petitioner_counsel_name = f.Str()
    petitioner_grant_date = f.Date()
    petitioner_patent_number = f.Str()
    petitioner_application_number_text = f.Str()

    # Appellant Information
    appellant_technology_center_number = f.Str()
    appellant_patent_owner_name = f.Str()
    appellant_party_name = f.Str()
    appellant_group_art_unit_number = f.Str()
    appellant_inventor_name = f.Str()
    appellant_counsel_name = f.Str()
    appellant_grant_date = f.Date()
    appellant_patent_number = f.Str()
    appellant_application_number_text = f.Str()
    appellant_publication_date = f.Date()
    appellant_publication_number = f.Str()
    third_party_name = f.Str()

    # Second Respondent (if any)
    second_respondent_party_name = f.Str()
    second_respondent_appl_number_text = f.Str()
    second_respondent_patent_number = f.Str()
    second_respondent_grant_date = f.Date()
    second_respondent_patent_owner_name = f.Str()
    second_respondent_inventor_name = f.Str()
    second_respondent_counsel_name = f.Str()
    second_respondent_g_a_u_number = f.Str()
    second_respondent_tech_center_number = f.Str()
    second_respondent_pub_number = f.Str()
    second_respondent_publication_date = f.Date()

    additional_respondents = f.List(AdditionalRespondentSchema, data_key="additionalRespondentPartyDataBag")


class PtabDocumentSchema(Schema):
    document_identifier = f.Str()
    document_category = f.Str()
    document_type_name = f.Str()
    document_number = f.Int()
    document_name = f.Str()
    document_filing_date = f.Date()
    document_title = f.Str()
    proceeding_number = f.Str()
    proceeding_type_category = f.Str()


class PtabDecisionSchema(Schema):
    proceeding_number = f.Str()
    board_rulings = f.List(f.Str())
    decision_type_category = f.Str()
    document_identifier = f.Str()
    document_name = f.Str()
    identifier = f.Str()
    issue_type = f.List(f.Str())
    object_uu_id = f.Str()
    petitioner_technology_center_number = f.Str()
    subdecision_type_category = f.Str()


class PtabDocumentPageSchema(Schema):
    num_found = f.Int(data_key="recordTotalQuantity")
    docs = f.List(PtabDocumentSchema, data_key="results")


class PtabProceedingPageSchema(Schema):
    num_found = f.Int(data_key="recordTotalQuantity")
    docs = f.List(PtabProceedingSchema, data_key="results")


class PtabDecisionPageSchema(Schema):
    num_found = f.Int(data_key="recordTotalQuantity")
    docs = f.List(PtabDecisionSchema, data_key="results")



import json
from pathlib import Path

import lxml.etree as ET
from patent_client.util.test import compare_dicts

from .schema import BiblioResultSchema
from .schema import ClaimsSchema
from .schema import DescriptionSchema
from .schema import ImagesSchema
from .schema import SearchSchema

test_dir = Path(__file__).parent / "test"
expected_dir = Path(__file__).parent / "test" / "expected"


def test_biblio():
    tree = ET.parse(test_dir / "biblio_example.xml")
    result = BiblioResultSchema().load(tree)
    assert len(result.documents) == 1
    d = result.documents[0]
    assert d.title == "Bi-directional steerable catheter control handle"
    assert d.country == "US"
    assert d.doc_number == "2006142694"
    assert d.family_id == "35840242"
    for f in (
        "publication_reference_docdb",
        "publication_reference_epodoc",
        "application_reference_docdb",
        "application_reference_epodoc",
        "application_reference_original",
    ):
        assert hasattr(d, f)
    assert len(d.citations) == 99


def test_claims():
    tree = ET.parse(test_dir / "claims_example.xml")
    result = ClaimsSchema().load(tree)
    example_file = expected_dir / "claims_example.json"
    example_file.write_text(result.to_json(indent=2))
    expected = json.loads(example_file.read_text())
    compare_dicts(json.loads(result.to_json()), expected)


def test_description():
    tree = ET.parse(test_dir / "description_example.xml")
    result = DescriptionSchema().load(tree)
    assert result.document_id.country == "EP"
    assert result.document_id.doc_number == "1000000"
    assert result.document_id.kind == "A1"
    assert len(result.description) == 10544


def test_search():
    tree = ET.parse(test_dir / "search_example.xml")
    result = SearchSchema().load(tree)
    assert result.query == 'applicant = "nine energy"'
    assert result.num_results == 15
    assert result.begin == 1
    assert result.end == 25
    assert len(result.results) == 15
    r = result.results[3]
    assert r.family_id == "70280448"
    assert r.id_type == "docdb"
    assert r.country == "US"


def test_images():
    tree = ET.parse(test_dir / "image_example.xml")
    result = ImagesSchema().load(tree)
    expected_file = expected_dir / "image_example.json"
    expected_file.write_text(result.to_json())
    expected = json.loads(expected_file.read_text())
    compare_dicts(json.loads(result.to_json()), expected)


def test_full_cycle():
    tree = ET.parse(test_dir / "full_cycle_example.xml")
    result = BiblioResultSchema().load(tree)
    assert len(result.documents) == 2
    d = result.documents[0]
    assert d.title == "Apparatus for manufacturing green bricks for the brick manufacturing industry"


def test_biblio_2():
    filename = "biblio_example_4"
    input = test_dir / f"{filename}.xml"
    tree = ET.parse(input)
    result = BiblioResultSchema().load(tree)
    expected_file = expected_dir / f"{filename}.json"
    # expected_file.write_text(result.to_json(indent=2))
    compare_dicts(json.loads(result.to_json()), json.loads(expected_file.read_text()))



from pathlib import Path

import pytest

from .api import PublishedAsyncApi

expected_dir = Path(__file__).parent / "test" / "expected"


class TestPublishedBiblioAsyncApi:
    @pytest.mark.asyncio
    async def test_doc_example_biblio(self):
        result = await PublishedAsyncApi.biblio.get_biblio("EP1000000.A1", format="epodoc")
        expected_file = expected_dir / "ep1000000_biblio_result.xml"
        expected_file.write_text(result, encoding="utf8")
        expected = expected_file.read_text(encoding="utf8")
        assert result == expected

    @pytest.mark.asyncio
    async def test_doc_example_full_cycle(self):
        result = await PublishedAsyncApi.biblio.get_full_cycle("EP1000000.A1", format="epodoc")
        expected_file = expected_dir / "ep1000000_full_cycle_result.xml"
        expected_file.write_text(result, encoding="utf8")
        expected = expected_file.read_text(encoding="utf8")
        assert result == expected

    @pytest.mark.asyncio
    async def test_doc_example_abstract(self):
        result = await PublishedAsyncApi.biblio.get_abstract("EP1000000.A1", format="epodoc")
        expected_file = expected_dir / "ep1000000_abstract_result.xml"
        expected_file.write_text(result, encoding="utf8")
        expected = expected_file.read_text(encoding="utf8")
        assert result == expected


class TestSearchApi:
    @pytest.mark.asyncio
    async def test_search(self):
        result = await PublishedAsyncApi.search.search("ti=plastic")
        expected_file = expected_dir / "search_result.xml"
        # expected_file.write_text(result)
        expected = expected_file.read_text()
        assert result == expected


class TestFullTextAsyncApi:
    @pytest.mark.asyncio
    async def test_description(self):
        result = await PublishedAsyncApi.fulltext.get_description("EP1000000.A1", format="epodoc")
        expected_file = expected_dir / "ep1000000_description_result.xml"
        # expected_file.write_text(result)
        expected = expected_file.read_text()
        assert result == expected

    @pytest.mark.asyncio
    async def test_claims(self):
        result = await PublishedAsyncApi.fulltext.get_claims("EP1000000.A1", format="epodoc")
        expected_file = expected_dir / "ep1000000_claims_result.xml"
        # expected_file.write_text(result)
        expected = expected_file.read_text()
        assert result == expected



import pytest

from ..session import asession
from ..session import OpsAuthenticationError
from .model.search import Inpadoc


class TestPublished:
    def test_inpadoc_manager(self):
        result = Inpadoc.objects.filter(applicant="Microsoft")
        assert len(result) > 20
        countries = list(result.limit(20).values_list("country", flat=True))
        assert sum(1 for c in countries if c == "US") >= 1

    def test_get_biblio_from_result(self):
        doc = Inpadoc.objects.filter(applicant="Google").first()
        result = doc.biblio
        assert result.title is not None

    def test_get_claims_from_result(self):
        result = Inpadoc.objects.get("WO2009085664A2")
        assert len(result.claims.claims) == 20
        assert len(result.claims.claim_text) == 4830

    def test_get_description_from_result(self):
        result = Inpadoc.objects.get("WO2009085664A2")
        assert len(result.description) == 47955

    def test_get_family_from_result(self):
        result = Inpadoc.objects.get("WO2009085664A2")
        assert len(result.family) >= 20

    def test_get_biblio_from_wo(self):
        result = Inpadoc.objects.get("WO2009085664A2").biblio
        assert result.abstract is not None

    def test_can_index_inpadoc_result(self):
        result = Inpadoc.objects.filter(applicant="Tesla")
        assert result[0] != result[1]

    def test_can_handle_single_item_ipc_classes(self):
        result = Inpadoc.objects.get("WO2020081771").biblio
        assert result.intl_class is not None

    def test_issue_41(self):
        result = Inpadoc.objects.get("JP2005533465A").biblio
        assert result.title == None

    def test_issue_76_no_credential(self):
        key = asession.key
        asession.key = None
        with pytest.raises(OpsAuthenticationError):
            pub = Inpadoc.objects.get("EP3082535A1")
        asession.key = key

    def test_issue_76_with_credential(self):
        pub = Inpadoc.objects.get("EP3082535A1")
        assert pub.biblio.title == "AUTOMATIC FLUID DISPENSER"


class TestPublished:
    @pytest.mark.asyncio
    async def test_inpadoc_manager(self):
        result = Inpadoc.objects.filter(applicant="Microsoft")
        assert await result.alen() > 20
        countries = await result.limit(20).values_list("country", flat=True).ato_list()
        assert sum(1 for c in countries if c == "US") >= 1

    @pytest.mark.asyncio
    async def test_get_biblio_from_result(self):
        doc = await Inpadoc.objects.filter(applicant="Google").afirst()
        result = doc.biblio
        assert result.title is not None

    @pytest.mark.asyncio
    async def test_get_claims_from_result(self):
        result = await Inpadoc.objects.aget("WO2009085664A2")
        assert len(result.claims.claims) == 20
        assert len(result.claims.claim_text) == 4830

    @pytest.mark.asyncio
    async def test_get_description_from_result(self):
        result = await Inpadoc.objects.aget("WO2009085664A2")
        assert len(result.description) == 47955

    @pytest.mark.asyncio
    async def test_get_family_from_result(self):
        result = await Inpadoc.objects.aget("WO2009085664A2")
        assert len(result.family) >= 20

    @pytest.mark.asyncio
    async def test_get_biblio_from_wo(self):
        result = await Inpadoc.objects.aget("WO2009085664A2")
        assert result.biblio.abstract is not None

    @pytest.mark.asyncio
    async def test_can_index_inpadoc_result(self):
        result = Inpadoc.objects.filter(applicant="Tesla")
        first = await result.afirst()
        second = await result.offset(1).afirst()
        assert first != second

    @pytest.mark.asyncio
    async def test_can_handle_single_item_ipc_classes(self):
        result = await Inpadoc.objects.aget("WO2020081771")
        assert result.biblio.intl_class is not None

    @pytest.mark.asyncio
    async def test_issue_41(self):
        result = await Inpadoc.objects.aget("JP2005533465A")
        assert result.biblio.title == None

    @pytest.mark.asyncio
    async def test_issue_76_with_credential(self):
        pub = await Inpadoc.objects.aget("EP3082535A1")
        assert pub.biblio.title == "AUTOMATIC FLUID DISPENSER"



import math
import typing as tp

from hishel._controller import Controller
from hishel._controller import get_age
from hishel._utils import BaseClock
from httpcore import Request
from httpcore import Response


def get_start_and_row_count(limit=None, offset=0, page_size=50):
    """Many API's have a start and row count parameter. This function
    converts an offset and limit into a iterator of start and row count
    tuples. If limit is None, then the iterator will be infinite.
    Otherwise, the iterator will be finite and the last row count may
    be less than the page size."""
    if not limit:
        page_no = 0
        while True:
            yield (page_no * page_size + offset, page_size)
            page_no += 1
    else:
        num_full_pages = math.floor(limit / page_size)
        last_page_size = limit % page_size
        for i in range(num_full_pages):
            yield (i * page_size + offset, page_size)
        if last_page_size:
            yield (num_full_pages * page_size + offset, last_page_size)


class SimpleController(Controller):
    def __init__(
        self,
        cacheable_methods: tp.Optional[tp.List[str]] = None,
        cacheable_status_codes: tp.Optional[tp.List[int]] = None,
        allow_heuristics: bool = False,
        clock: tp.Optional[BaseClock] = None,
        allow_stale: bool = False,
        always_revalidate: bool = False,
        max_age: tp.Optional[int] = None,
    ):
        super().__init__(
            cacheable_methods, cacheable_status_codes, allow_heuristics, clock, allow_stale, always_revalidate
        )
        self._max_age = max_age

    def construct_response_from_cache(
        self, request: Request, response: Response, original_request: Request
    ) -> Response | Request | None:
        # Use of responses with status codes 301 and 308 is always
        # legal as long as they don't adhere to any caching rules.
        if response.status in (301, 308):
            return response
        age = get_age(response, self._clock)
        is_fresh = age < self._max_age
        if is_fresh:
            return response
        else:
            # Otherwise, make a conditional request
            self._make_request_conditional(request=request, response=response)
            return request



from copy import deepcopy
from typing import Generic

import httpx
import inflection
from patent_client.util import Manager
from patent_client.util import ModelType
from patent_client.util.request_util import get_start_and_row_count

from . import schema_doc
from .api import PtabApi
from .model import PtabDecision
from .model import PtabDocument
from .model import PtabProceeding
from .util import peds_to_ptab

asession = httpx.AsyncClient(verify=False)


class PtabManager(Manager, Generic[ModelType]):
    url = "https://developer.uspto.gov/ptab-api"
    page_size = 25
    instance_schema = None
    api_method = None

    async def _aget_results(self):
        query = deepcopy(self.config.filter)
        query = peds_to_ptab(query)
        query["sort"] = " ".join(inflection.camelize(o, uppercase_first_letter=False) for o in self.config.order_by)
        for start, rows in get_start_and_row_count(self.config.limit, self.config.offset, self.page_size):
            query["start"] = start
            query["rows"] = rows
            page = await self.api_method(**query)
            for doc in page.docs:
                yield doc
            if len(page.docs) < rows:
                break

    async def alen(self):
        page = await self.api_method(**peds_to_ptab(self.config.filter))
        return min(self.config.limit, page.num_found) if self.config.limit else page.num_found

    def allowed_filters(self):
        params = schema_doc["paths"][self.path]["get"]["parameters"]
        return {inflection.underscore(p["name"]): p["description"] for p in params}


class PtabProceedingManager(PtabManager[PtabProceeding]):
    path = "/proceedings"
    primary_key = "proceeding_number"
    api_method = PtabApi.get_proceedings


class PtabDocumentManager(PtabManager[PtabDocument]):
    path = "/documents"
    primary_key = "document_identifier"
    api_method = PtabApi.get_documents


class PtabDecisionManager(PtabManager[PtabDecision]):
    path = "/decisions"
    primary_key = "identifier"
    api_method = PtabApi.get_decisions



# flake8: noqa
# nopycln: file
import time

from .version import __version__  # noqa

start = time.time()

import nest_asyncio

nest_asyncio.apply()

from pathlib import Path
from .settings import load_settings

SETTINGS = load_settings()

import logging
import logging.handlers

# Revert base directory to local if there's an access problem

BASE_DIR = Path(SETTINGS.DEFAULT.BASE_DIR).expanduser()
try:
    BASE_DIR.mkdir(exist_ok=True, parents=True)
except OSError:
    BASE_DIR = Path(__file__).parent.parent.parent / "_build"
    BASE_DIR.mkdir(exist_ok=True, parents=True)
    SETTINGS.DEFAULT.BASE_DIR = str(BASE_DIR)

LOG_FILENAME = BASE_DIR / SETTINGS.DEFAULT.LOG_FILE

# Set up a specific logger with our desired output level
logger = logging.getLogger(__name__)
logger.setLevel(SETTINGS.DEFAULT.LOG_LEVEL)


# Add the log message handler to the logger
handler = logging.FileHandler(LOG_FILENAME)
handler.setFormatter(logging.Formatter("%(asctime)s:%(levelname)s:%(name)s:%(message)s"))
logger.addHandler(handler)

logger.info(f"Starting Patent Client with log level {SETTINGS.DEFAULT.LOG_LEVEL}")


# Set up yankee
import yankee

yankee.use_model = True


from patent_client.epo.ops.published.model import Inpadoc  # isort:skip

# from patent_client.usitc.model import ITCAttachment
# from patent_client.usitc.model import ITCDocument
# from patent_client.usitc.model import ITCInvestigation
from patent_client.uspto.assignment.model import Assignment  # isort:skip
from patent_client.uspto.peds.model import USApplication  # isort:skip
from patent_client.uspto.ptab.model import PtabDecision  # isort:skip
from patent_client.uspto.ptab.model import PtabDocument  # isort:skip
from patent_client.uspto.ptab.model import PtabProceeding  # isort:skip
from patent_client.uspto.global_dossier.model import (
    GlobalDossier,
    GlobalDossierApplication,
)  # isort:skip
from patent_client.uspto.public_search.model import (
    PublicSearch,
    PublicSearchDocument,
    Patent,
    PatentBiblio,
    PublishedApplication,
    PublishedApplicationBiblio,
)  # isort:skip

elapsed = time.time() - start
logger.debug(f"Startup Complete!, took {elapsed:.3f} seconds")

cache_dir = Path("~/.patent_client/cache").expanduser()
cache_dir.parent.mkdir(exist_ok=True, parents=True)

__all__ = [
    "Inpadoc",
    "Assignment",
    "PublicSearch",
    "PublicSearchDocument",
    "Patent",
    "PatentBiblio",
    "PublishedApplication",
    "PublishedApplicationBiblio",
    "USApplication",
    "PtabDecision",
    "PtabDocument",
    "PtabProceeding",
    "GlobalDossier",
]



from .biblio import BiblioResult
from .biblio import InpadocBiblio
from .fulltext import Claims
from .fulltext import Description
from .images import ImageDocument
from .images import Images
from .search import Inpadoc
from .search import Search

__all__ = [
    "BiblioResult",
    "InpadocBiblio",
    "Claims",
    "Description",
    "Images",
    "ImageDocument",
    "Inpadoc",
    "Search",
]



import dataclasses
import datetime
from collections import abc


def resolve_list(item, key):
    item_list = resolve(item, key)

    if item_list is None:
        return []

    if isinstance(item_list, list):
        return item_list

    return [item_list]


def resolve_attribute(obj, key):
    if isinstance(obj, abc.Mapping):
        obj = obj[key]
    else:
        obj = getattr(obj, key)
    if callable(obj):
        obj = obj()
    return obj


def resolve(item, key):
    if key is None:
        return item
    accessors = key.split(".")
    try:
        for accessor in accessors:
            if isinstance(item, abc.Sequence) and accessor.isdigit():
                item = item[int(accessor)]
            elif isinstance(item, abc.Sequence) and not accessor.isdigit():
                item = [resolve_attribute(i, accessor) for i in item]
            else:
                item = resolve_attribute(item, accessor)
    except Exception as e:
        return None
    return item


def to_dict(obj, item_class=dict, collection_class=list):
    # print(f"Obj is {obj}")
    if isinstance(obj, abc.Mapping):
        # print(f"Casting as Mapping with class {item_class}")
        return item_class((k, to_dict(v, item_class, collection_class, convert_dates)) for k, v in obj.items())
    elif dataclasses.is_dataclass(obj):
        # print("Converting Dataclass")
        return to_dict(item_class(obj), item_class, collection_class, convert_dates)
    elif isinstance(obj, abc.Iterable) and not isinstance(obj, (str, bytes)):
        # print(f"Casting as Collection with {collection_class}")
        return collection_class(to_dict(i, item_class, collection_class, convert_dates) for i in obj)
    elif convert_dates and isinstance(obj, datetime.date):
        return datetime.datetime.combine(obj, datetime.datetime.min.time())
    else:
        # print("No cast - passing through")
        return obj



import datetime as dt

import hishel
from patent_client import SETTINGS
from patent_client.util.asyncio_util import SyncProxy

NS = {
    "http://ops.epo.org": None,
    "http://www.epo.org/exchange": None,
    "http://www.epo.org/fulltext": None,
    "http://www.epo.org/register": None,
}

import logging

logger = logging.getLogger(__name__)


class OpsAuthenticationError(Exception):
    pass


class OpsForbiddenError(Exception):
    pass


class OpsAsyncSession(hishel.AsyncCacheClient):
    def __init__(self, *args, key=None, secret=None, **kwargs):
        super(OpsAsyncSession, self).__init__(*args, **kwargs)
        self.key: str = key
        self.secret: str = secret
        self.expires: dt.datetime = dt.datetime.utcnow()
        self.sync = SyncProxy(self)

    async def request(self, *args, **kwargs):
        response = await super(OpsAsyncSession, self).request(*args, **kwargs)
        if response.status_code in (403, 400):
            auth_response = await self.get_token()
            response = await super(OpsAsyncSession, self).request(*args, **kwargs)
        return response

    async def get_token(self):
        auth_url = "https://ops.epo.org/3.2/auth/accesstoken"
        response = await super(OpsAsyncSession, self).request(
            "post",
            auth_url,
            auth=(self.key, self.secret),
            data={"grant_type": "client_credentials"},
        )
        if response.status_code == 401:
            logger.debug(f"EPO Authentication Error!\n{response.text}")
            raise OpsAuthenticationError(
                "Failed to authenticate with EPO OPS! Please check your credentials. See the setup instructions at https://patent-client.readthedocs.io/en/stable/getting_started.html"
            )
        elif response.status_code == 403:
            logger.debug(f"EPO Forbidden Error\n{response.text}")
            raise OpsForbiddenError("Your EPO Request Failed - Quota Exceeded / Blacklisted / Blocked")
        response.raise_for_status()

        data = response.json()
        self.expires = dt.datetime.fromtimestamp(int(data["issued_at"]) / 1000) + dt.timedelta(
            seconds=int(data["expires_in"])
        )
        self.headers["Authorization"] = f"Bearer {data['access_token']}"
        return response


asession = OpsAsyncSession(key=SETTINGS.EPO.API_KEY, secret=SETTINGS.EPO.API_SECRET, follow_redirects=True)



from __future__ import annotations

import datetime
import warnings
from dataclasses import dataclass
from dataclasses import field
from typing import *

from patent_client import session
from patent_client.util import Model
from patent_client.util.base.related import get_model
from yankee.data import ListCollection


@dataclass
class AssignmentPage:
    num_found: int
    docs: "List[Assignment]" = field(default_factory=ListCollection)


@dataclass
class Assignment(Model):
    __manager__ = "patent_client.uspto.assignment.manager.AssignmentManager"
    id: str
    conveyance_text: str
    last_update_date: str
    page_count: int
    recorded_date: datetime.date
    corr_name: str = None
    corr_address: str = None
    assignors: "ListCollection[Assignor]" = field(default_factory=ListCollection)
    assignees: "ListCollection[Assignee]" = field(default_factory=ListCollection)
    properties: "ListCollection[Property]" = field(repr=False, default_factory=ListCollection)
    """Properties objects associated with this Assignment"""
    assignment_record_has_images: bool = False
    transaction_date: "Optional[datetime.date]" = None
    date_produced: "Optional[datetime.date]" = None

    @property
    def image_url(self) -> str:
        reel, frame = self.id.split("-")
        reel = reel.rjust(6, "0")
        frame = frame.rjust(4, "0")
        return f"http://legacy-assignments.uspto.gov/assignments/assignment-pat-{reel}-{frame}.pdf"

    def download(self):
        """downloads the PDF associated with the assignment to the current working directory"""
        response = session.get(self.image_url, stream=True)
        with open(f"{self.id}.pdf", "wb") as f:
            f.write(response.raw.read())


@dataclass
class Property(Model):
    appl_id: "Optional[str]" = None
    invention_title: "Optional[str]" = None
    inventors: "Optional[str]" = None
    # Numbers
    pct_num: "Optional[str]" = None
    intl_reg_num: "Optional[str]" = None
    publ_num: "Optional[str]" = None
    pat_num: "Optional[str]" = None
    # Dates
    filing_date: "Optional[datetime.date]" = None
    intl_publ_date: "Optional[datetime.date]" = None
    issue_date: "Optional[datetime.date]" = None
    publ_date: "Optional[datetime.date]" = None

    def __repr__(self):
        return f"Property(appl_id={self.appl_id}, invention_title={self.invention_title})"

    @property
    def application(self) -> "patent_client.uspto.peds.model.USApplication":
        """The related US Application"""
        try:
            appl_id = getattr(self, "appl_id", None)
            if appl_id is not None:
                return get_model("patent_client.uspto.peds.model.USApplication").objects.get(appl_id=appl_id)
            appl_id = getattr(self, "pct_num", None)
            if appl_id is not None:
                return get_model("patent_client.uspto.peds.model.USApplication").objects.get(appl_id=appl_id)
            pub_num = getattr(self, "publ_num", None)
            if pub_num is not None:
                return get_model("patent_client.uspto.peds.model.USApplication").objects.get(
                    app_early_pub_number=pub_num
                )
            pat_num = getattr(self, "pat_num", None)
            if pat_num is not None:
                return get_model("patent_client.uspto.peds.model.USApplication").objects.get(patent_number=pat_num)
        except Exception:
            pass
        warnings.warn(f"Unable to find application for {self}")
        return None

    @property
    def patent(self) -> "patent_client.uspto.fulltext.patent.model.Patent":
        """The related US Patent, if any"""
        return get_model("patent_client.uspto.peds.fulltext.patent.model.Patent").objects.get(
            publication_number=self.pat_num
        )

    @property
    def publication(
        self,
    ) -> "patent_client.uspto.fulltext.published_application.model.PublishedApplication":
        """The related US Publication, if any"""
        return get_model("patent_client.uspto.peds.fulltext.patent.model.PublishedApplication").objects.get(
            publication_number=self.publ_num
        )


@dataclass
class Assignor(Model):
    name: str
    ex_date: datetime.date
    date_ack: datetime.datetime = None


@dataclass
class Assignee(Model):
    name: str
    address: "Optional[str]" = None



conversions = {
    "appl_id": "application_number",
    "inventor": "inventor_name",
}


def peds_to_ptab(query: dict) -> dict:
    for k, v in conversions.items():
        if k in query:
            query[v] = query.pop(k)
    return query



import os
from collections import defaultdict
from pathlib import Path

import yaml
from yankee.util import AttrDict

DEFAULT_SETTINGS_FILE = Path(__file__).parent / "default_settings.yml"
DEFAULT_SETTINGS = AttrDict.convert(yaml.safe_load(DEFAULT_SETTINGS_FILE.read_text()))


def load_settings_from_env():
    settings = defaultdict(dict)
    for k, v in os.environ.items():
        if not k.startswith("PATENT_CLIENT_"):
            continue
        k = k.replace("PATENT_CLIENT_", "")
        if "__" in k:
            section, k = k.split("__")
        else:
            section = "DEFAULT"
        settings[section][k] = v
    return settings


def load_user_settings():
    file = Path("~/.patent_client_config.yaml").expanduser()
    if file.exists():
        return yaml.safe_load(file.read_text())
    else:
        file.write_text(DEFAULT_SETTINGS_FILE.read_text())
        return dict()


def merge_settings(*settings):
    output = defaultdict(dict)
    for s in settings:
        for section, values in s.items():
            output[section] = {**output[section], **values}
    return output


def load_settings():
    return AttrDict.convert(
        merge_settings(
            DEFAULT_SETTINGS,
            load_user_settings(),
            load_settings_from_env(),
        )
    )



import datetime
import re
from dataclasses import dataclass
from dataclasses import field
from fileinput import filename
from pathlib import Path
from typing import *

from patent_client import session
from patent_client.util import Model
from yankee.data import ListCollection

from ...util.base.related import get_model


@dataclass
class AdditionalRespondent(Model):
    application_number_text: str = None
    inventor_name: str = None
    patent_number: str = None
    party_name: str = None


@dataclass
class PtabProceeding(Model):
    """A PTAB Proceeding - e.g. IPR/CBM/DER Trial, Patent Appeal, Interference, etc.

    All fields are query-able. Date ranges can be formed by inserting "from" or "to" on a query
    for a date range.

    """

    __manager__ = "patent_client.uspto.ptab.manager.PtabProceedingManager"
    # Proceeding Metadata
    last_modified_date: "datetime.datetime" = None
    last_modified_user_id: "datetime.datetime" = None
    institution_decision_date: "datetime.date" = None
    proceeding_filing_date: "datetime.date" = None
    accorded_filing_date: "datetime.date" = None
    proceeding_status_category: str = None
    proceeding_number: str = None
    proceeding_last_modified_date: "datetime.date" = None
    proceeding_type_category: str = None
    subproceeding_type_category: str = None
    decision_date: "datetime.date" = None
    docket_notice_mail_date: "datetime.date" = None
    declaration_date: "datetime.date" = None
    style_name_text: str = None

    # Respondent Information
    respondent_technology_center_number: str = None
    respondent_patent_owner_name: str = None
    respondent_party_name: str = None
    respondent_group_art_unit_number: str = None
    respondent_inventor_name: str = None
    respondent_counsel_name: str = None
    respondent_grant_date: "datetime.date" = None
    respondent_patent_number: str = None
    respondent_application_number_text: str = None
    respondent_publication_number: str = None
    respondent_publication_date: "datetime.date" = None

    # Petitioner Information
    petitioner_technology_center_number: str = None
    petitioner_patent_owner_name: str = None
    petitioner_party_name: str = None
    petitioner_group_art_unit_number: str = None
    petitioner_inventor_name: str = None
    petitioner_counsel_name: str = None
    petitioner_grant_date: "datetime.date" = None
    petitioner_patent_number: str = None
    petitioner_application_number_text: str = None

    # Appellant Information
    appellant_technology_center_number: str = None
    appellant_patent_owner_name: str = None
    appellant_party_name: str = None
    appellant_group_art_unit_number: str = None
    appellant_inventor_name: str = None
    appellant_counsel_name: str = None
    appellant_grant_date: "datetime.date" = None
    appellant_patent_number: str = None
    appellant_application_number_text: str = None
    appellant_publication_date: "datetime.date" = None
    appellant_publication_number: str = None
    third_party_name: str = None

    # Second Respondent (if any)
    second_respondent_party_name: str = None
    second_respondent_appl_number_text: str = None
    second_respondent_patent_number: str = None
    second_respondent_grant_date: "datetime.date" = None
    second_respondent_patent_owner_name: str = None
    second_respondent_inventor_name: str = None
    second_respondent_counsel_name: str = None
    second_respondent_g_a_u_number: str = None
    second_respondent_tech_center_number: str = None
    second_respondent_pub_number: str = None
    second_respondent_publication_date: "datetime.date" = None

    additional_respondents: "ListCollection[str]" = field(default_factory=list)

    def __repr__(self):
        return f"PtabProceeding(subproceeding_type_category='{self.subproceeding_type_category}', proceeding_number='{self.proceeding_number}', proceeding_status_category='{self.proceeding_status_category}', proceeding_type_category='{self.proceeding_type_category}', respondent_party_name='{self.respondent_party_name}')"

    @property
    def documents(
        self,
    ) -> "ListCollection[patent_client.uspto.ptab.model.PtabDocument]":
        """Documents associated with the Proceeding"""
        return get_model("patent_client.uspto.ptab.model.PtabDocument").objects.filter(
            proceeding_number=self.proceeding_number
        )

    @property
    def decisions(
        self,
    ) -> "ListCollection[patent_client.uspto.ptab.model.PtabDecision]":
        """Decisions associated with the Proceeding"""
        return get_model("patent_client.uspto.ptab.model.PtabDecision").objects.filter(
            proceeding_number=self.proceeding_number
        )

    @property
    def us_application(
        self,
    ) -> "ListCollection[patent_client.uspto.peds.model.USApplication]":
        """The US Application provided by PEDS associated with the Proceeding"""
        return get_model("patent_client.uspto.peds.model.USApplication").objects.get(
            patent_number=self.respondent_patent_number
        )


fname_re = re.compile(r"[<>:\"/\|?*]")


@dataclass
class PtabDocument(Model):
    __manager__ = "patent_client.uspto.ptab.manager.PtabDocumentManager"
    document_identifier: str = field(repr=False)
    document_category: "Optional[str]" = None
    document_type_name: "Optional[str]" = None
    document_number: "Optional[int]" = None
    document_name: "Optional[str]" = None
    document_filing_date: "Optional[datetime.date]" = None
    proceeding_number: "Optional[str]" = field(repr=False, default=None)
    proceeding_type_category: "Optional[str]" = field(repr=False, default=None)
    title: "Optional[str]" = None

    @property
    def proceeding(self) -> "patent_client.uspto.ptab.model.PtabProceeding":
        """The PTAB proceeding associated with the document"""
        return get_model("patent_client.uspto.ptab.model.PtabProceeding").objects.get(
            proceeding_number=self.proceeding_number
        )

    def download(self, path="."):
        name, ext = self.document_name.rsplit(".", 1)
        name = name[:100] + "." + ext
        filename = f"[{str(self.document_number).rjust(4, '0')}] {self.document_filing_date.isoformat()} - {name}"
        filename = filename.encode(encoding="ascii", errors="ignore").decode("ascii")
        filename = fname_re.sub("", filename)
        out_path = Path(path)
        if out_path.is_dir():
            out_path = Path(path) / filename
        else:
            out_path = out_path
        if out_path.exists():
            return out_path
        with session.get(
            f"https://developer.uspto.gov/ptab-api/documents/{self.document_identifier}/download",
            verify=False,
        ) as r:
            r.raise_for_status()
            with out_path.open("wb") as f:
                for chunk in r.iter_content(chunk_size=8192):
                    f.write(chunk)
        return out_path


@dataclass
class PtabDecision(Model):
    __manager__ = "patent_client.uspto.ptab.manager.PtabDecisionManager"
    proceeding_number: str
    board_rulings: "List[str]" = field(default_factory=ListCollection)
    decision_type_category: str = None
    document_identifier: str = None
    document_name: str = None
    identifier: str = None
    subdecision_type_category: str = None
    issue_type: "Optional[str]" = None
    object_uu_id: "Optional[str]" = None
    petitioner_technology_center_number: "Optional[str]" = None

    @property
    def proceeding(self) -> "patent_client.uspto.ptab.model.PtabProceeding":
        """The PTAB proceeding associated with the document"""
        return get_model("patent_client.uspto.ptab.model.PtabProceeding").objects.get(
            proceeding_number=self.proceeding_number
        )


@dataclass
class PtabDocumentPage(Model):
    __manager__ = "patent_client.uspto.ptab.manager.PtabDocumentManager"
    num_found: int = None
    docs: ListCollection[PtabDocument] = field(default_factory=ListCollection)


@dataclass
class PtabProceedingPage(Model):
    __manager__ = "patent_client.uspto.ptab.manager.PtabDocumentManager"
    num_found: int = None
    docs: ListCollection[PtabDocument] = field(default_factory=ListCollection)


@dataclass
class PtabDecisionPage(Model):
    __manager__ = "patent_client.uspto.ptab.manager.PtabDocumentManager"
    num_found: int = None
    docs: ListCollection[PtabDocument] = field(default_factory=ListCollection)



import json
import logging
from datetime import date
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import AsyncIterator

import inflection
from patent_client.util.asyncio_util import run_sync
from patent_client.util.base.manager import Manager
from patent_client.util.request_util import get_start_and_row_count
from PyPDF2 import PdfFileMerger

from .api import PatentExaminationDataSystemApi
from .model import USApplication
from .schema import DocumentSchema
from .schema import USApplicationSchema
from .session import session


logger = logging.getLogger(__name__)


class HttpException(Exception):
    pass


class USApplicationManager(Manager[USApplication]):
    primary_key = "appl_id"
    query_url = "https://ped.uspto.gov/api/queries"
    page_size = 20
    __schema__ = USApplicationSchema()

    def __init__(self, *args, **kwargs):
        super(USApplicationManager, self).__init__(*args, **kwargs)
        self.pages = dict()

    async def alen(self):
        api = PatentExaminationDataSystemApi()
        max_length = (await api.create_query(**self.get_query_params())).num_found
        return min(max_length, self.config.limit) if self.config.limit else max_length

    async def _aget_results(self) -> AsyncIterator[USApplication]:
        query_params = self.get_query_params()
        api = PatentExaminationDataSystemApi()
        for start, rows in get_start_and_row_count(self.config.limit, self.config.offset, self.page_size):
            page = await api.create_query(**{**query_params, "start": start, "rows": rows})
            for app in page.applications:
                yield app
            if len(page.applications) < rows:
                break

    def get_query_params(self):
        if "query" in self.config.filter:
            query_text = self.config.filter["query"]
        else:
            query = list()
            for k, v in self.config.filter.items():
                field = inflection.camelize(k, uppercase_first_letter=False)
                if not v:
                    continue
                elif type(v) in (list, tuple):
                    v = (str(i) for i in v)
                    body = " OR ".join(f'"{value}"' if " " in value else value for value in v)
                else:
                    body = v
                query.append(f"{field}:({body})")
            query_text = " AND ".join(query).strip()
        if self.config.order_by:
            sort_query = ""
            for s in self.config.order_by:
                if s[0] == "-":
                    sort_query += f"{inflection.camelize(s[1:], uppercase_first_letter=False)} desc ".strip()
                else:
                    sort_query += (f"{inflection.camelize(s, uppercase_first_letter=False)} asc").strip()
        else:
            sort_query = None

        mm = "0%" if "appEarlyPubNumber" not in query else "90%"

        query_data = {
            "query": query_text,
            "facet": False,
            "minimum_match": mm,
        }
        if sort_query:
            query_data["sort"] = sort_query
        return query_data

    @property
    def allowed_filters(self):
        fields = self.fields()
        return list(fields.keys())

    def fields(self):
        """List of fields available to the API"""
        if not hasattr(self.__class__, "_fields"):
            url = "https://ped.uspto.gov/api/search-fields"
            response = run_sync(session.get(url))
            if not response.status_code == 200:
                raise ValueError("Can't get fields!")
            raw = response.json()
            output = {inflection.underscore(key): value for (key, value) in raw.items()}
            self.__class__._fields = output
        return self.__class__._fields

    def is_online(self):
        return run_sync(PatentExaminationDataSystemApi.is_online)

    @property
    def query_fields(self):
        fields = self.fields()
        for k in sorted(fields.keys()):
            if "facet" in k:
                continue
            print(f"{k} ({fields[k]})")


class DateEncoder(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, date):
            return o.isoformat()
        return json.JSONEncoder.default(self, o)


class DocumentManager(Manager):
    query_url = "https://ped.uspto.gov/api/queries/cms/public/"
    __schema__ = DocumentSchema()

    async def alen(self):
        url = self.query_url + self.config.filter["appl_id"]
        response = await session.get(url)
        response.raise_for_status()
        return len(response.json())

    async def _aget_results(self) -> AsyncIterator[USApplication]:
        url = self.query_url + self.config.filter["appl_id"]
        response = await session.get(url)
        for item in response.json():
            yield self.__schema__.load(item)

    def download(self, docs, path="."):
        run_sync(self.adownload(docs, path))

    async def adownload(self, docs, path="."):
        if str(path)[-4:].lower() == ".pdf":
            # If we've been given a specific filename, use it
            out_file = Path(path)
        else:
            out_file = Path(path) / "package.pdf"

        files = list()
        try:
            with TemporaryDirectory() as tmpdir:
                for doc in docs:
                    if doc.access_level_category == "PUBLIC":
                        files.append((await doc.adownload(tmpdir), doc))

                out_pdf = PdfFileMerger()
                page = 0
                for f, doc in files:
                    bookmark = f"{doc.mail_room_date} - {doc.code} - {doc.description}"
                    out_pdf.append(str(f), bookmark=bookmark, import_bookmarks=False)
                    page += doc.page_count

                out_pdf.write(str(out_file))
        except (PermissionError, NotADirectoryError):
            # This is needed due to a bug in Windows that prevents cleanup of the tmpdir
            pass
        return out_file



from patent_client.epo.ops.number_service.schema import DocumentIdSchema
from patent_client.epo.ops.published.schema.biblio import DocDbNumberField
from patent_client.epo.ops.util import Schema
from yankee.xml import fields as f
from yankee.xml.schema import fields as f

from ..util import Schema


class PriorityClaimSchema(Schema):
    application_number = DocDbNumberField(".//epo:document-id")
    application_reference = DocumentIdSchema(".//epo:document-id")
    sequence = f.Int("./@sequence")
    kind = f.Str("./@kind")
    active = f.Bool(".//epo:priority-active-indicator", true_value="YES")


class FamilyMemberSchema(Schema):
    publication_number = DocDbNumberField('.//epo:publication-reference/epo:document-id[@document-id-type="docdb"]')
    application_number = DocDbNumberField('.//epo:application-reference/epo:document-id[@document-id-type="docdb"]')
    family_id = f.Str("./@family-id")
    publication_reference = f.List(DocumentIdSchema, ".//epo:publication-reference/epo:document-id")
    application_reference = f.List(DocumentIdSchema, ".//epo:application-reference/epo:document-id")
    priority_claims = f.List(PriorityClaimSchema, ".//epo:priority-claim")


class FamilySchema(Schema):
    publication_reference = DocumentIdSchema(".//ops:patent-family/ops:publication-reference")
    num_records = f.Int(".//ops:patent-family/@total-result-count")
    publication_number = DocDbNumberField(
        './/ops:patent-family/ops:publication-reference/epo:document-id[@document-id-type="docdb"]'
    )
    family_members = f.List(FamilyMemberSchema, ".//ops:family-member")



from .date_parse import parse_date  # noqa: F401
from .date_parse import parse_datetime  # noqa: F401
from .date_parse import parse_duration  # noqa: F401
from .date_parse import parse_time  # noqa: F401
